##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""
import copy
import reportlab.lib.styles
import reportlab.lib.enums
import reportlab.platypus
from z3c.rml import attr, directive, interfaces, paraparser, occurence, special
from z3c.rml import SampleStyleSheet


class IInitialize(interfaces.IRMLDirectiveSignature):
    """Do some RML processing initialization."""
    occurence.containing(
        occurence.ZeroOrMore('name', special.IName),
        occurence.ZeroOrMore('alias', special.IAlias),
        )

class Initialize(directive.RMLDirective):
    signature = IInitialize
    factories = {
        'name': special.Name,
        'alias': special.Alias,
        }


class ISpanStyle(interfaces.IRMLDirectiveSignature):
    """Defines a span style and gives it a name."""

    name = attr.String(
        title=u'Name',
        description=u'The name of the style.',
        required=True)

    alias = attr.String(
        title=u'Alias',
        description=u'An alias under which the style will also be known as.',
        required=False)

    parent = attr.Style(
        title=u'Parent',
        description=(u'The apragraph style that will be used as a base for '
                     u'this one.'),
        required=False)

    fontName = attr.String(
        title=u'Font Name',
        description=u'The name of the font for the span.',
        required=False)

    fontSize = attr.Measurement(
        title=u'Font Size',
        description=u'The font size for the text of the span.',
        required=False)

    textColor = attr.Color(
        title=u'Text Color',
        description=u'The color in which the text will appear.',
        required=False)

    backColor = attr.Color(
        title=u'Background Color',
        description=u'The background color of the span.',
        required=False)


class SpanStyle(directive.RMLDirective):
    signature = ISpanStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop('parent', paraparser.SpanStyle('DefaultSpan'))
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class IBaseParagraphStyle(ISpanStyle):

    leading = attr.Measurement(
        title=u'Leading',
        description=(u'The height of a single paragraph line. It includes '
                     u'character height.'),
        required=False)

    leftIndent = attr.Measurement(
        title=u'Left Indentation',
        description=u'General indentation on the left side.',
        required=False)

    rightIndent = attr.Measurement(
        title=u'Right Indentation',
        description=u'General indentation on the right side.',
        required=False)

    firstLineIndent = attr.Measurement(
        title=u'First Line Indentation',
        description=u'The indentation of the first line in the paragraph.',
        required=False)

    alignment = attr.Choice(
        title=u'Alignment',
        description=u'The text alignment.',
        choices=interfaces.ALIGN_CHOICES,
        required=False)

    spaceBefore = attr.Measurement(
        title=u'Space Before',
        description=u'The vertical space before the paragraph.',
        required=False)

    spaceAfter = attr.Measurement(
        title=u'Space After',
        description=u'The vertical space after the paragraph.',
        required=False)

    bulletFontName = attr.String(
        title=u'Bullet Font Name',
        description=u'The font in which the bullet character will be rendered.',
        required=False)

    bulletFontSize = attr.Measurement(
        title=u'Bullet Font Size',
        description=u'The font size of the bullet character.',
        required=False)

    bulletIndent = attr.Measurement(
        title=u'Bullet Indentation',
        description=u'The indentation that is kept for a bullet point.',
        required=False)

    wordWrap = attr.String(
        title=u'Word Wrap Method',
        description=(u'When set to "CJK", invoke CJK word wrapping'),
        required=False)

    borderWidth = attr.Measurement(
        title=u'Paragraph Border Width',
        description=u'The width of the paragraph border.',
        required=False)

    borderPadding = attr.Padding(
        title=u'Paragraph Border Padding',
        description=u'Padding of the paragraph.',
        required=False)

    borderColor = attr.Color(
        title=u'Border Color',
        description=u'The color in which the paragraph border will appear.',
        required=False)

    borderRadius = attr.Measurement(
        title=u'Paragraph Border Radius',
        description=u'The radius of the paragraph border.',
        required=False)

    allowWidows = attr.Boolean(
        title=u'Allow Widows',
        description=(u'Allow widows.'),
        required=False)

    allowOrphans = attr.Boolean(
        title=u'Allow Orphans',
        description=(u'Allow orphans.'),
        required=False)

    textTransforms = attr.Choice(
        title=u'Text Transforms',
        description=u'Text transformations.',
        choices=interfaces.TEXT_TRANSFORM_CHOICES,
        required=False)

    endDots = attr.String(
        title=u'End Dots',
        description=u'Characters/Dots at the end of a paragraph.',
        required=False)

    splitLongWords = attr.Boolean(
        title=u'Split Long Words',
        description=(u'Try to split long words at the end of a line.'),
        default=True,
        required=False)

    underlineProportion = attr.Float(
        title=u'Underline Proportion',
        description=(u'When set to non-zero value, it provides the scaling '
                     u'factor for the width of the underine based on '
                     u'current font size.'),
        min=0.0,
        default=0.0,
        required=False)

    bulletAnchor = attr.Choice(
        title=u'Bullet Anchor',
        description=u'The place at which the bullet is anchored.',
        choices=interfaces.BULLET_ANCHOR_CHOICES,
        default='start',
        required=False)

    # Attributes not part of the official style attributes, but are accessed
    # by the paragraph renderer.

    keepWithNext = attr.Boolean(
        title=u'Keep with Next',
        description=(u'When set, this paragraph will always be in the same '
                     u'frame as the following flowable.'),
        required=False)

    pageBreakBefore = attr.Boolean(
        title=u'Page Break Before',
        description=(u'Specifies whether a page break should be inserted '
                     u'before the directive.'),
        required=False)

    frameBreakBefore = attr.Boolean(
        title=u'Frame Break Before',
        description=(u'Specifies whether a frame break should be inserted '
                     u'before the directive.'),
        required=False)


class IParagraphStyle(IBaseParagraphStyle):
    """Defines a paragraph style and gives it a name."""

    name = attr.String(
        title=u'Name',
        description=u'The name of the style.',
        required=True)

    alias = attr.String(
        title=u'Alias',
        description=u'An alias under which the style will also be known as.',
        required=False)

    parent = attr.Style(
        title=u'Parent',
        description=(u'The apragraph style that will be used as a base for '
                     u'this one.'),
        required=False)

class ParagraphStyle(directive.RMLDirective):
    signature = IParagraphStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop(
            'parent', SampleStyleSheet['Normal'])
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class ITableStyleCommand(interfaces.IRMLDirectiveSignature):

    start = attr.Sequence(
        title=u'Start Coordinates',
        description=u'The start table coordinates for the style instruction',
        value_type=attr.Combination(
            value_types=(attr.Integer(),
                         attr.Choice(choices=interfaces.SPLIT_CHOICES))
            ),
        default=[0, 0],
        min_length=2,
        max_length=2,
        required=True)

    stop = attr.Sequence(
        title=u'End Coordinates',
        description=u'The end table coordinates for the style instruction',
        value_type=attr.Combination(
            value_types=(attr.Integer(),
                         attr.Choice(choices=interfaces.SPLIT_CHOICES))
            ),
        default=[-1, -1],
        min_length=2,
        max_length=2,
        required=True)

class TableStyleCommand(directive.RMLDirective):
    name = None

    def process(self):
        args = [self.name]
        args += self.getAttributeValues(valuesOnly=True)
        self.parent.style.add(*args)


class IBlockFont(ITableStyleCommand):
    """Set the font properties for the texts."""

    name = attr.String(
        title=u'Font Name',
        description=u'The name of the font for the cell.',
        required=False)

    size = attr.Measurement(
        title=u'Font Size',
        description=u'The font size for the text of the cell.',
        required=False)

    leading = attr.Measurement(
        title=u'Leading',
        description=(u'The height of a single text line. It includes '
                     u'character height.'),
        required=False)

class BlockFont(TableStyleCommand):
    signature = IBlockFont
    name = 'FONT'

class IBlockLeading(ITableStyleCommand):
    """Set the text leading."""

    length = attr.Measurement(
        title=u'Length',
        description=(u'The height of a single text line. It includes '
                     u'character height.'),
        required=True)

class BlockLeading(TableStyleCommand):
    signature = IBlockLeading
    name = 'LEADING'

class IBlockTextColor(ITableStyleCommand):
    """Set the text color."""

    colorName = attr.Color(
        title=u'Color Name',
        description=u'The color in which the text will appear.',
        required=True)

class BlockTextColor(TableStyleCommand):
    signature = IBlockTextColor
    name = 'TEXTCOLOR'

class IBlockAlignment(ITableStyleCommand):
    """Set the text alignment."""

    value = attr.Choice(
        title=u'Text Alignment',
        description=u'The text alignment within the cell.',
        choices=interfaces.ALIGN_TEXT_CHOICES,
        required=True)

class BlockAlignment(TableStyleCommand):
    signature = IBlockAlignment
    name = 'ALIGNMENT'

class IBlockLeftPadding(ITableStyleCommand):
    """Set the left padding of the cells."""

    length = attr.Measurement(
        title=u'Length',
        description=u'The size of the padding.',
        required=True)

class BlockLeftPadding(TableStyleCommand):
    signature = IBlockLeftPadding
    name = 'LEFTPADDING'

class IBlockRightPadding(ITableStyleCommand):
    """Set the right padding of the cells."""

    length = attr.Measurement(
        title=u'Length',
        description=u'The size of the padding.',
        required=True)

class BlockRightPadding(TableStyleCommand):
    signature = IBlockRightPadding
    name = 'RIGHTPADDING'

class IBlockBottomPadding(ITableStyleCommand):
    """Set the bottom padding of the cells."""

    length = attr.Measurement(
        title=u'Length',
        description=u'The size of the padding.',
        required=True)

class BlockBottomPadding(TableStyleCommand):
    signature = IBlockBottomPadding
    name = 'BOTTOMPADDING'

class IBlockTopPadding(ITableStyleCommand):
    """Set the top padding of the cells."""

    length = attr.Measurement(
        title=u'Length',
        description=u'The size of the padding.',
        required=True)

class BlockTopPadding(TableStyleCommand):
    signature = IBlockTopPadding
    name = 'TOPPADDING'

class IBlockBackground(ITableStyleCommand):
    """Define the background color of the cells.

    It also supports alternating colors.
    """

    colorName = attr.Color(
        title=u'Color Name',
        description=u'The color to use as the background for every cell.',
        required=False)

    colorsByRow = attr.Sequence(
        title=u'Colors By Row',
        description=u'A list of colors to be used circularly for rows.',
        value_type=attr.Color(acceptNone=True),
        required=False)

    colorsByCol = attr.Sequence(
        title=u'Colors By Column',
        description=u'A list of colors to be used circularly for columns.',
        value_type=attr.Color(acceptNone=True),
        required=False)

class BlockBackground(TableStyleCommand):
    signature = IBlockBackground
    name = 'BACKGROUND'

    def process(self):
        args = [self.name]
        if 'colorsByRow' in self.element.keys():
            args = [BlockRowBackground.name]
        elif 'colorsByCol' in self.element.keys():
            args = [BlockColBackground.name]

        args += self.getAttributeValues(valuesOnly=True)
        self.parent.style.add(*args)

class IBlockRowBackground(ITableStyleCommand):
    """Define the background colors for rows."""

    colorNames = attr.Sequence(
        title=u'Colors By Row',
        description=u'A list of colors to be used circularly for rows.',
        value_type=attr.Color(),
        required=True)

class BlockRowBackground(TableStyleCommand):
    signature = IBlockRowBackground
    name = 'ROWBACKGROUNDS'

class IBlockColBackground(ITableStyleCommand):
    """Define the background colors for columns."""

    colorNames = attr.Sequence(
        title=u'Colors By Row',
        description=u'A list of colors to be used circularly for rows.',
        value_type=attr.Color(),
        required=True)

class BlockColBackground(TableStyleCommand):
    signature = IBlockColBackground
    name = 'COLBACKGROUNDS'

class IBlockValign(ITableStyleCommand):
    """Define the vertical alignment of the cells."""

    value = attr.Choice(
        title=u'Vertical Alignment',
        description=u'The vertical alignment of the text with the cells.',
        choices=interfaces.VALIGN_TEXT_CHOICES,
        required=True)

class BlockValign(TableStyleCommand):
    signature = IBlockValign
    name = 'VALIGN'

class IBlockSpan(ITableStyleCommand):
    """Define a span over multiple cells (rows and columns)."""

class BlockSpan(TableStyleCommand):
    signature = IBlockSpan
    name = 'SPAN'

class ILineStyle(ITableStyleCommand):
    """Define the border line style of each cell."""

    kind = attr.Choice(
        title=u'Kind',
        description=u'The kind of line actions to be taken.',
        choices=('GRID', 'BOX', 'OUTLINE', 'INNERGRID',
                 'LINEBELOW', 'LINEABOVE', 'LINEBEFORE', 'LINEAFTER'),
        required=True)

    thickness = attr.Measurement(
        title=u'Thickness',
        description=u'Line Thickness',
        default=1,
        required=True)

    colorName = attr.Color(
        title=u'Color',
        description=u'The color of the border line.',
        default=None,
        required=True)

    cap = attr.Choice(
        title=u'Cap',
        description=u'The cap at the end of a border line.',
        choices=interfaces.CAP_CHOICES,
        default=1,
        required=True)

    dash = attr.Sequence(
        title=u'Dash-Pattern',
        description=u'The dash-pattern of a line.',
        value_type=attr.Measurement(),
        default=None,
        required=False)

    join = attr.Choice(
        title=u'Join',
        description=u'The way lines are joined together.',
        choices=interfaces.JOIN_CHOICES,
        default=1,
        required=False)

    count = attr.Integer(
        title=u'Count',
        description=(u'Describes whether the line is a single (1) or '
                     u'double (2) line.'),
        default=1,
        required=False)

class LineStyle(TableStyleCommand):
    signature = ILineStyle

    def process(self):
        name = self.getAttributeValues(select=('kind',), valuesOnly=True)[0]
        args = [name]
        args += self.getAttributeValues(ignore=('kind',), valuesOnly=True,
                                        includeMissing=True)
        self.parent.style.add(*args)

class IBlockTableStyle(interfaces.IRMLDirectiveSignature):
    """A style defining the look of a table."""
    occurence.containing(
        occurence.ZeroOrMore('blockFont', IBlockFont),
        occurence.ZeroOrMore('blockLeading', IBlockLeading),
        occurence.ZeroOrMore('blockTextColor', IBlockTextColor),
        occurence.ZeroOrMore('blockAlignment', IBlockAlignment),
        occurence.ZeroOrMore('blockLeftPadding', IBlockLeftPadding),
        occurence.ZeroOrMore('blockRightPadding', IBlockRightPadding),
        occurence.ZeroOrMore('blockBottomPadding', IBlockBottomPadding),
        occurence.ZeroOrMore('blockTopPadding', IBlockTopPadding),
        occurence.ZeroOrMore('blockBackground', IBlockBackground),
        occurence.ZeroOrMore('blockRowBackground', IBlockRowBackground),
        occurence.ZeroOrMore('blockColBackground', IBlockColBackground),
        occurence.ZeroOrMore('blockValign', IBlockValign),
        occurence.ZeroOrMore('blockSpan', IBlockSpan),
        occurence.ZeroOrMore('lineStyle', ILineStyle)
        )

    id = attr.String(
        title=u'Id',
        description=u'The name/id of the style.',
        required=True)

    keepWithNext = attr.Boolean(
        title=u'Keep with Next',
        description=(u'When set, this paragraph will always be in the same '
                     u'frame as the following flowable.'),
        required=False)

class BlockTableStyle(directive.RMLDirective):
    signature = IBlockTableStyle

    factories = {
        'blockFont': BlockFont,
        'blockLeading': BlockLeading,
        'blockTextColor': BlockTextColor,
        'blockAlignment': BlockAlignment,
        'blockLeftPadding': BlockLeftPadding,
        'blockRightPadding': BlockRightPadding,
        'blockBottomPadding': BlockBottomPadding,
        'blockTopPadding': BlockTopPadding,
        'blockBackground': BlockBackground,
        'blockRowBackground': BlockRowBackground,
        'blockColBackground': BlockColBackground,
        'blockValign': BlockValign,
        'blockSpan': BlockSpan,
        'lineStyle': LineStyle,
        }

    def process(self):
        kw = dict(self.getAttributeValues())
        id  = kw.pop('id')
        # Create Style
        self.style = reportlab.platypus.tables.TableStyle()
        for name, value in kw.items():
            setattr(self.style, name, value)
        # Fill style
        self.processSubDirectives()
        # Add style to the manager
        manager = attr.getManager(self)
        manager.styles[id] = self.style


class IMinimalListStyle(interfaces.IRMLDirectiveSignature):

    leftIndent = attr.Measurement(
        title=u'Left Indentation',
        description=u'General indentation on the left side.',
        required=False)

    rightIndent = attr.Measurement(
        title=u'Right Indentation',
        description=u'General indentation on the right side.',
        required=False)

    bulletColor = attr.Color(
        title=u'Bullet Color',
        description=u'The color in which the bullet will appear.',
        required=False)

    bulletFontName = attr.String(
        title=u'Bullet Font Name',
        description=u'The font in which the bullet character will be rendered.',
        required=False)

    bulletFontSize = attr.Measurement(
        title=u'Bullet Font Size',
        description=u'The font size of the bullet character.',
        required=False)

    bulletOffsetY = attr.Measurement(
        title=u'Bullet Y-Offset',
        description=u'The vertical offset of the bullet.',
        required=False)

    bulletDedent = attr.StringOrInt(
        title=u'Bullet Dedent',
        description=u'Either pixels of dedent or auto (default).',
        required=False)

    bulletDir = attr.Choice(
        title=u'Bullet Layout Direction',
        description=u'The layout direction of the bullet.',
        choices=('ltr', 'rtl'),
        required=False)

    bulletFormat = attr.String(
        title=u'Bullet Format',
        description=u'A formatting expression for the bullet text.',
        required=False)

    bulletType = attr.Choice(
        title=u'Bullet Type',
        description=u'The type of number to display.',
        choices=interfaces.ORDERED_LIST_TYPES + \
                interfaces.UNORDERED_BULLET_VALUES,
        doLower=False,
        required=False)

class IBaseListStyle(IMinimalListStyle):

    start = attr.Combination(
        title=u'Start Value',
        description=u'The counter start value.',
        value_types=(
            # Numeric start value.
            attr.Integer(),
            # Bullet types.
            attr.Choice(choices=interfaces.UNORDERED_BULLET_VALUES),
            # Arbitrary text.
            attr.String(),
            ),
        required=False)


class IListStyle(IBaseListStyle):
    """Defines a list style and gives it a name."""

    name = attr.String(
        title=u'Name',
        description=u'The name of the style.',
        required=True)

    parent = attr.Style(
        title=u'Parent',
        description=(u'The list style that will be used as a base for '
                     u'this one.'),
        required=False)


class ListStyle(directive.RMLDirective):
    signature = IListStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop(
            'parent', reportlab.lib.styles.ListStyle(name='List'))
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class IStylesheet(interfaces.IRMLDirectiveSignature):
    """A styleheet defines the styles that can be used in the document."""
    occurence.containing(
        occurence.ZeroOrOne('initialize', IInitialize),
        occurence.ZeroOrMore('spanStyle', ISpanStyle),
        occurence.ZeroOrMore('paraStyle', IParagraphStyle),
        occurence.ZeroOrMore('blockTableStyle', IBlockTableStyle),
        occurence.ZeroOrMore('listStyle', IListStyle),
        # TODO:
        #occurence.ZeroOrMore('boxStyle', IBoxStyle),
        )

class Stylesheet(directive.RMLDirective):
    signature = IStylesheet

    factories = {
        'initialize': Initialize,
        'spanStyle': SpanStyle,
        'paraStyle': ParagraphStyle,
        'blockTableStyle': BlockTableStyle,
        'listStyle': ListStyle,
        }



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML Attribute Implementation
"""
import collections
import logging
import os
import re
import six
import reportlab.graphics.widgets.markers
import reportlab.lib.colors
import reportlab.lib.pagesizes
import reportlab.lib.styles
import reportlab.lib.units
import reportlab.lib.utils
import zope.interface
import zope.schema
from lxml import etree
from importlib import import_module
from z3c.rml import interfaces, SampleStyleSheet

MISSING = object()
logger = logging.getLogger("z3c.rml")


def getFileInfo(directive):
    root = directive
    while root.parent:
        root = root.parent
    # Elements added through API won't have a sourceline
    if directive.element.sourceline is None:
        return '(file %s)' % (root.filename)

    return '(file %s, line %i)' % (
        root.filename, directive.element.sourceline)


def getManager(context, interface=None):
    if interface is None:
        # Avoid circular imports
        from z3c.rml import interfaces
        interface = interfaces.IManager
    # Walk up the path until the manager is found
    # Using interface.providedBy is much slower because it does many more checks
    while (
        context is not None and
        not interface in context.__class__.__dict__.get('__implemented__', {})
    ):
        context = context.parent
    # If no manager was found, raise an error
    if context is None:
        raise ValueError('The manager could not be found.')
    return context


def deprecated(oldName, attr, reason):
    zope.interface.directlyProvides(attr, interfaces.IDeprecated)
    attr.deprecatedName = oldName
    attr.deprecatedReason = reason
    return attr


class RMLAttribute(zope.schema.Field):
    """An attribute of the RML directive."""

    missing_value = MISSING
    default = MISSING

    def fromUnicode(self, ustr):
        """See zope.schema.interfaces.IField"""
        if self.context is None:
            raise ValueError('Attribute not bound to a context.')
        return super(RMLAttribute, self).fromUnicode(six.text_type(ustr))

    def get(self):
        """See zope.schema.interfaces.IField"""
        # If the attribute has a deprecated partner and the deprecated name
        # has been specified, use it.
        if (interfaces.IDeprecated.providedBy(self) and
            self.deprecatedName in self.context.element.attrib):
            name = self.deprecatedName
            logger.warning(
                u'Deprecated attribute "%s": %s %s' % (
                name, self.deprecatedReason, getFileInfo(self.context)))
        else:
            name = self.__name__
        # Extract the value.
        value = self.context.element.get(name, self.missing_value)
        # Get the correct default value.
        if value is self.missing_value:
            if self.default is not None:
                return self.default
            return self.missing_value
        return self.fromUnicode(value)


class BaseChoice(RMLAttribute):
    choices = {}
    doLower = True

    def fromUnicode(self, value):
        if self.doLower:
            value = value.lower()
        if value in self.choices:
            return self.choices[value]
        raise ValueError(
            '%r not a valid value for attribute "%s". %s' % (
            value, self.__name__, getFileInfo(self.context)))


class Combination(RMLAttribute):
    """A combination of several other attribute types."""

    def __init__(self, value_types=(), *args, **kw):
        super(Combination, self).__init__(*args, **kw)
        self.value_types = value_types

    @property
    def element(self):
        return self.context.element

    @property
    def parent(self):
        return self.context.parent

    def fromUnicode(self, value):
        for value_type in self.value_types:
            bound = value_type.bind(self)
            try:
                return bound.fromUnicode(value)
            except ValueError:
                pass
        raise ValueError(
            '"%s" is not a valid value. %s' %(
                value, getFileInfo(self.context)))


class Text(RMLAttribute, zope.schema.Text):
    """A simple unicode string."""


if six.PY2:
    class String(RMLAttribute, zope.schema.Bytes):
        """A simple Bytes string."""
else:
    class String(Text):
        """Formerly a simple Bytes string, now the same as Text."""


class Integer(RMLAttribute, zope.schema.Int):
    """An integer. A minimum and maximum value can be specified."""
    # By making min and max simple attributes, we avoid some validation
    # problems.
    min = None
    max = None


class Float(RMLAttribute, zope.schema.Float):
    """An flaoting point. A minimum and maximum value can be specified."""
    # By making min and max simple attributes, we avoid some validation
    # problems.
    min = None
    max = None


class StringOrInt(RMLAttribute):
    """A (bytes) string or an integer."""

    def fromUnicode(self, value):
        try:
            return int(value)
        except ValueError:
            return str(value)


class Sequence(RMLAttribute, zope.schema._field.AbstractCollection):
    """A list of values of a specified type."""

    splitre = re.compile('[ \t\n,;]+')

    def __init__(self, splitre=None, *args, **kw):
        super(Sequence, self).__init__(*args, **kw)
        if splitre is not None:
            self.splitre = splitre

    def fromUnicode(self, ustr):
        if ustr.startswith('(') and ustr.endswith(')'):
            ustr = ustr[1:-1]
        ustr = ustr.strip()
        raw_values = self.splitre.split(ustr)
        result = [self.value_type.bind(self.context).fromUnicode(raw.strip())
                  for raw in raw_values]
        if ((self.min_length is not None and len(result) < self.min_length) or
            (self.max_length is not None and len(result) > self.max_length)):
            raise ValueError(
                'Length of sequence must be at least %s and at most %i. %s' % (
                self.min_length, self.max_length,
                getFileInfo(self.context)))
        return result


class IntegerSequence(Sequence):
    """A sequence of integers."""

    def __init__(
            self,
            numberingStartsAt=0,
            lowerBoundInclusive=True,
            upperBoundInclusive=True,
            *args, **kw):
        super(Sequence, self).__init__(*args, **kw)
        self.numberingStartsAt = numberingStartsAt
        self.lowerBoundInclusive = lowerBoundInclusive
        self.upperBoundInclusive = upperBoundInclusive

    def fromUnicode(self, ustr):
        ustr = ustr.strip()
        pieces = self.splitre.split(ustr)
        numbers = []
        for piece in pieces:
            # Ignore empty pieces.
            if not piece:
                continue
            # The piece is a range.
            if '-' in piece:
                start, end = map(int, piece.split('-'))
                # Make sure internally numbering starts as 0
                start -= self.numberingStartsAt
                end -= self.numberingStartsAt
                # Apply lower-bound exclusive.
                if not self.lowerBoundInclusive:
                    start += 1
                # Apply upper-bound inclusive.
                if self.upperBoundInclusive:
                    end += 1
                # Make range lower and upper bound inclusive.
                numbers.append((start, end))
                continue
            # The piece is just a number
            value = int(piece)
            value -= self.numberingStartsAt
            numbers.append((value, value+1))
        return numbers

class Choice(BaseChoice):
    """A choice of several values. The values are always case-insensitive."""

    def __init__(self, choices=None, doLower=True, *args, **kw):
        super(Choice, self).__init__(*args, **kw)
        if not isinstance(choices, dict):
            choices = collections.OrderedDict(
                [(val.lower() if doLower else val, val) for val in choices])
        else:
            choices = collections.OrderedDict(
                [(key.lower() if doLower else key, val)
                 for key, val in choices.items()])
        self.choices = choices
        self.doLower = doLower


class Boolean(BaseChoice):
    '''A boolean value.

    For true the values "true", "yes", and "1" are allowed. For false, the
    values "false", "no", "1" are allowed.
    '''
    choices = {'true': True, 'false': False,
               'yes': True, 'no': False,
               '1': True, '0': False,
               }

class TextBoolean(BaseChoice):
    '''A boolean value as text.

    ReportLab sometimes exposes low-level APIs, so we have to provide values
    that are directly inserted into the PDF.

    For "true" the values "true", "yes", and "1" are allowed. For "false", the
    values "false", "no", "1" are allowed.
    '''
    choices = {'true': 'true', 'false': 'false',
               'yes': 'true', 'no': 'false',
               '1': 'true', '0': 'false',
               }


class BooleanWithDefault(Boolean):
    '''This is a boolean field that can also receive the value "default".'''
    choices = Boolean.choices.copy()
    choices.update({'default': None})


class Measurement(RMLAttribute):
    '''This field represents a length value.

    The units "in" (inch), "cm", "mm", and "pt" are allowed. If no units are
    specified, the value is given in points/pixels.
    '''
    def __init__(self, allowPercentage=False, allowStar=False, *args, **kw):
        super(Measurement, self).__init__(*args, **kw)
        self.allowPercentage = allowPercentage
        self.allowStar = allowStar

    units = [
        (re.compile('^(-?[0-9\.]+)\s*in$'), reportlab.lib.units.inch),
        (re.compile('^(-?[0-9\.]+)\s*cm$'), reportlab.lib.units.cm),
        (re.compile('^(-?[0-9\.]+)\s*mm$'), reportlab.lib.units.mm),
        (re.compile('^(-?[0-9\.]+)\s*pt$'), 1),
        (re.compile('^(-?[0-9\.]+)\s*$'), 1)
    ]

    allowPercentage = False
    allowStar = False

    def fromUnicode(self, value):
        if value == 'None':
            return None
        if value == '*' and self.allowStar:
            return value
        if value.endswith('%') and self.allowPercentage:
            return value
        for unit in self.units:
            res = unit[0].search(value, 0)
            if res:
                return unit[1]*float(res.group(1))
        raise ValueError(
            'The value %r is not a valid measurement. %s' % (
            value, getFileInfo(self.context)))


class File(Text):
    """This field will return a file object.

    The value itself can eith be be a relative or absolute path. Additionally
    the following syntax is supported: [path.to.python.mpackage]/path/to/file
    """
    packageExtract = re.compile('^\[([0-9A-z_.]*)\]/(.*)$')

    doNotOpen = False

    def __init__(self, doNotOpen=False, *args, **kw):
        super(File, self).__init__(*args, **kw)
        self.doNotOpen = doNotOpen

    def fromUnicode(self, value):
        # Check whether the value is of the form:
        #    [<module.path>]/rel/path/image.gif"
        if value.startswith('['):
            result = self.packageExtract.match(value)
            if result is None:
                raise ValueError(
                    'The package-path-pair you specified was incorrect. %s' %(
                    getFileInfo(self.context)))
            modulepath, path = result.groups()
            module = import_module(modulepath)
            if six.PY2:
                value = os.path.join(os.path.dirname(module.__file__), path)
            else:
                # PEP 420 namespace support means that a module can have
                # multiple paths
                for module_path in module.__path__:
                    value = os.path.join(module_path, path)
                    if os.path.exists(value):
                        break
        # Under Python 3 all platforms need a protocol for local files
        if not six.moves.urllib.parse.urlparse(value).scheme:
            value = 'file:///' + os.path.abspath(value)
        # If the file is not to be opened, simply return the path.
        if self.doNotOpen:
            return value
        # Open/Download the file
        # We can't use urlopen directly because it has problems with data URIs
        # in 2.7 and 3.3 which are resolved in 3.4. Fortunately Reportlab
        # already has a utility function to help us work around this issue.
        fileObj = reportlab.lib.utils.open_for_read(value)
        sio = six.BytesIO(fileObj.read())
        fileObj.close()
        sio.seek(0)
        return sio


class Image(File):
    """Similar to the file File attribute, except that an image is internally
    expected."""

    def __init__(self, onlyOpen=False, *args, **kw):
        super(Image, self).__init__(*args, **kw)
        self.onlyOpen = onlyOpen

    def fromUnicode(self, value):
        if value.lower().endswith('.svg') or value.lower().endswith('.svgz'):
            return self._load_svg(value)
        fileObj = super(Image, self).fromUnicode(value)
        if self.onlyOpen:
            return fileObj
        return reportlab.lib.utils.ImageReader(fileObj)

    def _load_svg(self, value):
        manager = getManager(self.context)

        width = self.context.element.get('width')
        if width is not None:
            width = Measurement().fromUnicode(width)
        height = self.context.element.get('height')
        if height is not None:
            height = Measurement().fromUnicode(height)
        preserve = self.context.element.get('preserveAspectRatio')
        if preserve is not None:
            preserve = Boolean().fromUnicode(preserve)

        cache_key = '%s-%sx%s-%s' % (value, width, height, preserve)
        if cache_key in manager.svgs:
            return manager.svgs[cache_key]

        from gzip import GzipFile
        from reportlab.graphics import renderPM
        from xml.etree import cElementTree
        from z3c.rml.svg2rlg import Renderer

        fileObj = super(Image, self).fromUnicode(value)
        svg = fileObj.getvalue()
        if svg[:2] == b'\037\213':
            svg = GzipFile(fileobj=fileObj).read()
        svg = cElementTree.fromstring(svg)
        svg = Renderer(value).render(svg)

        if preserve:
            if width is not None or height is not None:
                if width is not None and height is None:
                    height = svg.height * width / svg.width
                elif height is not None and width is None:
                    width = svg.width * height / svg.height
                elif float(width) / height > float(svg.width) / svg.height:
                    width = svg.width * height / svg.height
                else:
                    height = svg.height * width / svg.width
        else:
            if width is None:
                width = svg.width
            if height is None:
                height = svg.height

        svg.scale(width / svg.width, height / svg.height)
        svg.width = width
        svg.height = height

        svg = renderPM.drawToPIL(svg, dpi=300)
        svg = reportlab.lib.utils.ImageReader(svg)
        svg.read = True # A hack to get ImageReader through as an open Image
                        # when used with imageAndFlowables
        manager.svgs[cache_key] = svg
        return svg


class Color(RMLAttribute):
    """Requires the input of a color. There are several supported formats.

    Three values in a row are interpreted as RGB value ranging from 0-255.
    A string is interpreted as a name to a pre-defined color.
    The 'CMYK()' wrapper around four values represents a CMYK color
    specification.
    """

    def __init__(self, acceptNone=False, acceptAuto=False, *args, **kw):
        super(Color, self).__init__(*args, **kw)
        self.acceptNone = acceptNone
        self.acceptAuto = acceptAuto

    def fromUnicode(self, value):
        if self.acceptNone and value.lower() == 'none':
            return None
        if self.acceptAuto and value.lower() == 'auto':
            return 'auto'
        manager = getManager(self.context)

        if value.startswith('rml:'):
            value = manager.get_name(value[4:], '#000000')

        if value in manager.colors:
            return manager.colors[value]
        try:
            return reportlab.lib.colors.toColor(value)
        # Bare except, since code raises string exception: Invalid color value
        except:
            raise ValueError(
                'The color specification "%s" is not valid. %s' % (
                value, getFileInfo(self.context)))

def _getStyle(context, value):
    manager = getManager(context)
    for styles in (manager.styles, SampleStyleSheet.byName):
        if value in styles:
            return styles[value]
        elif 'style.' + value in styles:
            return styles['style.' + value]
        elif value.startswith('style.') and value[6:] in styles:
            return styles[value[6:]]
    raise ValueError('Style %r could not be found. %s' % (
        value, getFileInfo(context)))

class Style(String):
    """Requires a valid style to be entered.

    Whether the style is a paragraph, table or box style is irrelevant, except
    that it has to fit the tag.
    """
    default = SampleStyleSheet.byName['Normal']

    def fromUnicode(self, value):
        return _getStyle(self.context, value)


class Padding(Sequence):
    """This attribute is specific for padding and will produce the proper
    length of the padding sequence."""

    def __init__(self, *args, **kw):
        kw.update(dict(value_type=Integer(), min_length=1, max_length=4))
        super(Padding, self).__init__(*args, **kw)

    def fromUnicode(self, value):
        seq = super(Padding, self).fromUnicode(value)
        # pdfgen does not like a single paddign value.
        if len(seq) == 1:
            seq.append(seq[0])
        return seq


class Symbol(Text):
    """This attribute should contain the text representation of a symbol to be
    used."""

    def fromUnicode(self, value):
        return reportlab.graphics.widgets.markers.makeMarker(value)


class PageSize(RMLAttribute):
    """A simple measurement pair that specifies the page size. Optionally you
    can also specify a the name of a page size, such as A4, letter, or legal.
    """

    sizePair = Sequence(value_type=Measurement())
    words = Sequence(value_type=String())

    def fromUnicode(self, value):
        # First try to get a pair
        try:
            return self.sizePair.bind(self.context).fromUnicode(value)
        except ValueError:
            pass
        # Now we try to lookup a name. The following type of combinations must
        # work: "Letter" "LETTER" "A4 landscape" "letter portrait"
        words = self.words.bind(self.context).fromUnicode(value)
        words = [word.lower() for word in words]
        # First look for the orientation
        orienter = None
        for orientation in ('landscape', 'portrait'):
            if orientation in words:
                orienter = getattr(reportlab.lib.pagesizes, orientation)
                words.remove(orientation)
        # We must have exactely one value left that matches a paper size
        pagesize = getattr(reportlab.lib.pagesizes, words[0].upper())
        # Now do the final touches
        if orienter:
            pagesize = orienter(pagesize)
        return pagesize


class TextNode(RMLAttribute):
    """Return the text content of an element."""

    def get(self):
        if self.context.element.text is None:
            return u''
        return six.text_type(self.context.element.text).strip()


class FirstLevelTextNode(TextNode):
    """Gets all the text content of an element without traversing into any
    child-elements."""

    def get(self):
        text = self.context.element.text or u''
        for child in self.context.element.getchildren():
            text += child.tail or u''
        return text.strip()


class TextNodeSequence(Sequence, TextNode):
    """A sequence of values retrieved from the element's content."""

    def get(self):
        return self.fromUnicode(self.context.element.text)


class TextNodeGrid(TextNodeSequence):
    """A grid/matrix of values retrieved from the element's content.

    The number of columns is specified for every case, but the number of rows
    is dynamic.
    """

    def __init__(self, columns=None, *args, **kw):
        super(TextNodeGrid, self).__init__(*args, **kw)
        self.columns = columns

    def fromUnicode(self, ustr):
        result = super(TextNodeGrid, self).fromUnicode(ustr)
        if len(result) % self.columns != 0:
            raise ValueError(
                'Number of elements must be divisible by %i. %s' %(
                self.columns, getFileInfo(self.context)))
        return [result[i*self.columns:(i+1)*self.columns]
                for i in range(len(result)//self.columns)]


class RawXMLContent(RMLAttribute):
    """Retrieve the raw content of an element.

    Only some special element substitution will be made.
    """

    def __init__(self, *args, **kw):
        super(RawXMLContent, self).__init__(*args, **kw)

    def get(self):
        # ReportLab's paragraph parser does not like attributes from other
        # namespaces; sigh. So we have to improvize.
        text = etree.tounicode(self.context.element, pretty_print=False)
        text = text[text.find('>')+1:text.rfind('<')]
        return text

class XMLContent(RawXMLContent):
    """Same as 'RawXMLContent', except that the whitespace is normalized."""

    def get(self):
        text = super(XMLContent, self).get()
        return text.strip().replace('\t', ' ')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Generate a DTD from the code
"""
import six
import zope.schema

from z3c.rml import attr, document, occurence

occurence2Symbol = {
    occurence.ZeroOrMore: '*',
    occurence.ZeroOrOne: '?',
    occurence.OneOrMore: '+',
    }


def generateElement(name, signature):
    if signature is None:
        return ''
    # Create the list of sub-elements.
    subElementList = []
    for occurence in signature.queryTaggedValue('directives', ()):
        subElementList.append(
            occurence.tag + occurence2Symbol.get(occurence.__class__, '')
            )
    fields = zope.schema.getFieldsInOrder(signature)
    for attrName, field in fields:
        if isinstance(field, attr.TextNode):
            subElementList.append('#PCDATA')
            break
    subElementList = ','.join(subElementList)
    if subElementList:
        subElementList = ' (' + subElementList + ')'
    text = '\n<!ELEMENT %s%s>' %(name, subElementList)
    # Create a list of attributes for this element.
    for attrName, field in fields:
        # Ignore text nodes, since they are not attributes.
        if isinstance(field, attr.TextNode):
            continue
        # Create the type
        if isinstance(field, attr.Choice):
            type = '(' + '|'.join(field.choices.keys()) + ')'
        else:
            type = 'CDATA'
        # Create required flag
        if field.required:
            required = '#REQUIRED'
        else:
            required = '#IMPLIED'
        # Put it all together
        text += '\n<!ATTLIST %s %s %s %s>' %(name, attrName, type, required)
    text += '\n'
    # Walk through all sub-elements, creating th eDTD entries for them.
    for occurence in signature.queryTaggedValue('directives', ()):
        text += generateElement(occurence.tag, occurence.signature)
    return text


def generate(useWrapper=False):
    text = generateElement('document', document.Document.signature)
    if useWrapper:
        text = '<!DOCTYPE RML [\n%s]>\n' %text
    return text

def main():
    six.print_(generate())



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``doc*`` directives.
"""
import reportlab.platypus
from z3c.rml import attr, directive, flowable, interfaces, occurence

class IDocAssign(interfaces.IRMLDirectiveSignature):
    """Assign a value to the namesapce."""

    var = attr.String(
        title=u'Variable Name',
        description=u'The name under which the value is stored.',
        required=True)

    expr = attr.String(
        title=u'Expression',
        description=u'The expression that creates the value when evaluated.',
        required=True)

class DocAssign(flowable.Flowable):
    signature = IDocAssign
    klass = reportlab.platypus.flowables.DocAssign


class IDocExec(interfaces.IRMLDirectiveSignature):
    """Execute a statement."""

    stmt = attr.String(
        title=u'Statement',
        description=u'The statement to be executed.',
        required=True)

class DocExec(flowable.Flowable):
    signature = IDocExec
    klass = reportlab.platypus.flowables.DocExec


class IDocPara(interfaces.IRMLDirectiveSignature):
    """Create a paragraph with the value returned from the expression."""

    expr = attr.String(
        title=u'Expression',
        description=u'The expression to be executed.',
        required=True)

    format = attr.String(
        title=u'Format',
        description=u'The format used to render the expression value.',
        required=False)

    style = attr.Style(
        title=u'Style',
        description=u'The style of the paragraph.',
        required=False)

    escape = attr.Boolean(
        title=u'Escape Text',
        description=u'When set (default) the expression value is escaped.',
        required=False)

class DocPara(flowable.Flowable):
    signature = IDocPara
    klass = reportlab.platypus.flowables.DocPara


class IDocAssert(interfaces.IRMLDirectiveSignature):
    """Assert a certain condition."""

    cond = attr.String(
        title=u'Condition',
        description=u'The condition to be asserted.',
        required=True)

    format = attr.String(
        title=u'Format',
        description=u'The text displayed if assertion fails.',
        required=False)

class DocAssert(flowable.Flowable):
    signature = IDocAssert
    klass = reportlab.platypus.flowables.DocAssert


class IDocElse(interfaces.IRMLDirectiveSignature):
    """Starts 'else' block."""

class DocElse(flowable.Flowable):
    signature = IDocElse

    def process(self):
        if not isinstance(self.parent, DocIf):
            raise ValueError("<docElse> can only be placed inside a <docIf>")
        self.parent.flow = self.parent.elseFlow


class IDocIf(flowable.IFlow):
    """Display story flow based on the value of the condition."""

    cond = attr.String(
        title=u'Condition',
        description=u'The condition to be tested.',
        required=True)

class DocIf(flowable.Flow):
    signature = IDocAssert
    klass = reportlab.platypus.flowables.DocIf

    def __init__(self, *args, **kw):
        super(flowable.Flow, self).__init__(*args, **kw)
        self.thenFlow = self.flow = []
        self.elseFlow = []

    def process(self):
        args = dict(self.getAttributeValues())
        self.processSubDirectives()
        dif = self.klass(
            thenBlock = self.thenFlow, elseBlock = self.elseFlow, **args)
        self.parent.flow.append(dif)

class IDocWhile(flowable.IFlow):
    """Repeat the included directives as long as the condition is true."""

    cond = attr.String(
        title=u'Condition',
        description=u'The condition to be tested.',
        required=True)

class DocWhile(flowable.Flow):
    signature = IDocAssert
    klass = reportlab.platypus.flowables.DocWhile

    def process(self):
        args = dict(self.getAttributeValues())
        self.processSubDirectives()
        dwhile = self.klass(whileBlock = self.flow, **args)
        self.parent.flow.append(dwhile)


flowable.Flow.factories['docAssign'] = DocAssign
flowable.Flow.factories['docExec'] = DocExec
flowable.Flow.factories['docPara'] = DocPara
flowable.Flow.factories['docAssert'] = DocAssert
flowable.Flow.factories['docIf'] = DocIf
flowable.Flow.factories['docElse'] = DocElse
flowable.Flow.factories['docWhile'] = DocWhile

flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('docAssign', IDocAssign),
     occurence.ZeroOrMore('docExec', IDocExec),
     occurence.ZeroOrMore('docPara', IDocPara),
     occurence.ZeroOrMore('docIf', IDocIf),
     occurence.ZeroOrMore('docElse', IDocElse),
     occurence.ZeroOrMore('docWhile', IDocWhile),
     )
    )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""
import six
import zope.interface
from reportlab import platypus
from z3c.rml import attr, directive, interfaces, occurence
from z3c.rml import canvas, flowable, page, stylesheet


class IStory(flowable.IFlow):
    """The story of the PDF file."""
    occurence.containing(
        *flowable.IFlow.getTaggedValue('directives'))

    firstPageTemplate = attr.Text(
        title=u'First Page Template',
        description=u'The first page template to be used.',
        default=None,
        required=False)

class Story(flowable.Flow):
    signature = IStory

    def process(self):
        self.parent.flowables = super(Story, self).process()
        self.parent.doc._firstPageTemplateIndex = self.getFirstPTIndex()

    def getFirstPTIndex(self):
        args = dict(self.getAttributeValues(select=('firstPageTemplate',)))
        fpt = args.pop('firstPageTemplate', None)
        if fpt is None:
            return 0
        for idx, pageTemplate in enumerate(self.parent.doc.pageTemplates):
            if pageTemplate.id == fpt:
                return idx
        raise ValueError('%r is not a correct page template id.' %fpt)


class IFrame(interfaces.IRMLDirectiveSignature):
    """A frame on a page."""

    x1 = attr.Measurement(
        title=u'X-Position',
        description=u'The X-Position of the lower-left corner of the frame.',
        allowPercentage=True,
        required=True)

    y1 = attr.Measurement(
        title=u'Y-Position',
        description=u'The Y-Position of the lower-left corner of the frame.',
        allowPercentage=True,
        required=True)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the frame.',
        allowPercentage=True,
        required=True)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the frame.',
        allowPercentage=True,
        required=True)

    id = attr.Text(
        title=u'Id',
        description=u'The id of the frame.',
        required=False)

    leftPadding = attr.Measurement(
        title=u'Left Padding',
        description=u'The left padding of the frame.',
        default=0,
        required=False)

    rightPadding = attr.Measurement(
        title=u'Right Padding',
        description=u'The right padding of the frame.',
        default=0,
        required=False)

    topPadding = attr.Measurement(
        title=u'Top Padding',
        description=u'The top padding of the frame.',
        default=0,
        required=False)

    bottomPadding = attr.Measurement(
        title=u'Bottom Padding',
        description=u'The bottom padding of the frame.',
        default=0,
        required=False)

    showBoundary = attr.Boolean(
        title=u'Show Boundary',
        description=u'A flag to show the boundary of the frame.',
        required=False)


class Frame(directive.RMLDirective):
    signature = IFrame

    def process(self):
        # get the page size
        size = self.parent.pt.pagesize
        if size is None:
            size = self.parent.parent.parent.doc.pagesize
        # Get the arguments
        args = dict(self.getAttributeValues())
        # Deal with percentages
        for name, dir in (('x1', 0), ('y1', 1), ('width', 0), ('height', 1)):
            if isinstance(args[name], six.string_types) and args[name].endswith('%'):
                args[name] = float(args[name][:-1])/100*size[dir]
        frame = platypus.Frame(**args)
        self.parent.frames.append(frame)


class IPageGraphics(canvas.IDrawing):
    """Define the page graphics for the page template."""

@zope.interface.implementer(interfaces.ICanvasManager)
class PageGraphics(directive.RMLDirective):
    signature = IPageGraphics

    def process(self):
        onPage = self.parent.pt.onPage

        def drawOnCanvas(canv, doc):
            onPage(canv, doc)
            canv.saveState()
            self.canvas = canv
            drawing = canvas.Drawing(self.element, self)
            drawing.process()
            canv.restoreState()

        self.parent.pt.onPage = drawOnCanvas


class Header(PageGraphics):

    def process(self):
        onPage = self.parent.pt.onPage

        def drawOnCanvas(canv, doc):
            onPage(canv, doc)
            canv.saveState()
            self.canvas = canv
            place = canvas.Place(self.element, self)
            place.process()
            canv.restoreState()

        self.parent.pt.onPage = drawOnCanvas


class Footer(Header):
    pass


class IPageTemplate(interfaces.IRMLDirectiveSignature):
    """Define a page template."""
    occurence.containing(
        occurence.OneOrMore('frame', IFrame),
        occurence.ZeroOrOne('pageGraphics', IPageGraphics),
        occurence.ZeroOrOne('mergePage', page.IMergePage),
        )

    id = attr.Text(
        title=u'Id',
        description=u'The id of the template.',
        required=True)

    pagesize = attr.PageSize(
        title=u'Page Size',
        description=u'The Page Size.',
        required=False)

    autoNextTemplate = attr.String(
        title=u'Auto Next Page Template',
        description=u'The page template to use automatically for the next page.',
        required=False)


class PageTemplate(directive.RMLDirective):
    signature = IPageTemplate
    attrMapping = {'autoNextTemplate': 'autoNextPageTemplate'}
    factories = {
        'frame': Frame,
        'pageGraphics': PageGraphics,
        'mergePage': page.MergePageInPageTemplate,
        'header': Header,
        'footer': Footer,
        }

    def process(self):
        args = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        pagesize = args.pop('pagesize', None)

        self.frames = []
        self.pt = platypus.PageTemplate(**args)

        self.processSubDirectives()
        self.pt.frames = self.frames

        if pagesize:
            self.pt.pagesize = pagesize

        self.parent.parent.doc.addPageTemplates(self.pt)


class ITemplate(interfaces.IRMLDirectiveSignature):
    """Define a page template."""
    occurence.containing(
        occurence.OneOrMore('pageTemplate', IPageTemplate),
        )

    pagesize = attr.PageSize(
        title=u'Page Size',
        description=u'The Page Size.',
        required=False)

    rotation = attr.Integer(
        title=u'Rotation',
        description=u'The rotation of the page in multiples of 90 degrees.',
        required=False)

    leftMargin = attr.Measurement(
        title=u'Left Margin',
        description=u'The left margin of the template.',
        default=0,
        required=False)

    rightMargin = attr.Measurement(
        title=u'Right Margin',
        description=u'The right margin of the template.',
        default=0,
        required=False)

    topMargin = attr.Measurement(
        title=u'Top Margin',
        description=u'The top margin of the template.',
        default=0,
        required=False)

    bottomMargin = attr.Measurement(
        title=u'Bottom Margin',
        description=u'The bottom margin of the template.',
        default=0,
        required=False)

    showBoundary = attr.Boolean(
        title=u'Show Boundary',
        description=u'A flag to show the boundary of the template.',
        required=False)

    allowSplitting = attr.Boolean(
        title=u'Allow Splitting',
        description=u'A flag to allow splitting over multiple templates.',
        required=False)

    title = attr.Text(
        title=u'Title',
        description=u'The title of the PDF document.',
        required=False)

    author = attr.Text(
        title=u'Author',
        description=u'The author of the PDF document.',
        required=False)


class Template(directive.RMLDirective):
    signature = ITemplate
    factories = {
        'pageTemplate': PageTemplate,
        }

    def process(self):
        args = self.getAttributeValues()
        args += self.parent.getAttributeValues(
            select=('debug', 'compression', 'invariant'),
            attrMapping={'debug': '_debug', 'compression': 'pageCompression'})
        args += (('cropMarks', self.parent.cropMarks),)

        self.parent.doc = platypus.BaseDocTemplate(
            self.parent.outputFile, **dict(args))
        self.processSubDirectives()



##############################################################################
#
# Copyright (c) 2007-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing all XML Locale functionality.
"""
import reportlab.platypus
import reportlab.lib.styles
import reportlab.graphics.widgets.markers
from reportlab.graphics import renderPDF, shapes
from reportlab.lib import colors
from z3c.rml import attr, interfaces, SampleStyleSheet


def myPreformatted(params):
    return reportlab.platypus.Preformatted('''
        Hey, this is a long text from a plugin. Hey, this is a long text from
        a plugin. Hey, this is a long text from a plugin. Hey, this is a long
        text from a plugin. Hey, this is a long text from a plugin. Hey, this
        is a long text from a plugin. Hey, this is a long text from a
        plugin. Hey, this is a long text from a plugin. Hey, this is a long
        text from a plugin. Hey, this is a long text from a plugin. Hey, this
        is a long text from a plugin. Hey, this is a long text from a
        plugin. Hey, this is a long text from a plugin. Hey, this is a long
        text from a plugin. Hey, this is a long text from a plugin. Hey, this
        is a long text from a plugin.''',
        SampleStyleSheet['Normal'])


class LinkURL(reportlab.platypus.flowables.Flowable):
    def __init__(self, link):
        self.link = link

    def wrap(self, *args):
        return (0, 0)

    def draw(self):
        self.canv.linkURL(self.link, None)


def linkURL(params):
    params = eval(params)
    return (
        reportlab.platypus.Paragraph(
            params[0], SampleStyleSheet['Normal']),
        LinkURL(*params))


class IMarker(interfaces.IRMLDirectiveSignature):
    x = attr.Measurement()
    y = attr.Measurement()
    dx = attr.Measurement()
    dy = attr.Measurement()
    size = attr.Measurement()
    fillColor = attr.Color(acceptNone=True, required=False)
    strokeColor = attr.Color(acceptNone=True, required=False)
    strokeWidth = attr.Measurement(required=False)
    arrowBarDx = attr.Measurement(required=False)
    arrowHeight = attr.Measurement(required=False)


def symbols(canvas, params):
    args = eval('dict(%s)' %params)
    name = args.pop('name')
    n = args.pop('n')
    for key, value in args.items():
        field = IMarker[key].bind(canvas.manager)
        args[key] = field.fromUnicode(value)
    m = reportlab.graphics.widgets.markers.makeMarker(name, **args)
    drawing = shapes.Drawing()
    drawing.add(m)
    for idx in range(n):
        drawing.drawOn(canvas, idx*args['dx'], idx*args['dy'])



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
svg2rlg is a tool to convert from SVG to reportlab graphics.

License : BSD

version 0.3
"""
import os
import re
import six
from gzip import GzipFile

from xml.etree import cElementTree

from reportlab.graphics import renderPDF

from reportlab.graphics.shapes import Drawing, Group, String, Line, Rect, Image
from reportlab.graphics.shapes import Circle, Ellipse, Polygon, PolyLine, Path
from reportlab.graphics.shapes import mmult

from reportlab.lib.units import pica, toLength
import reportlab.lib.colors as colors

class SVGError(Exception):
    pass

EOF = object()

class Lexer:
    """
    This style of implementation was inspired by this article:

        http://www.gooli.org/blog/a-simple-lexer-in-python/
    """
    Float = r'[-\+]?(?:(?:\d*\.\d+)|(?:\d+\.)|(?:\d+))(?:[Ee][-\+]?\d+)?'
    Int = r'[-\+]?\d+'

    lexicon = None
    ignore = None
    callbacks = None

    def __init__(self):
        lexicon = self.lexicon

        # create internal names for group matches
        groupnames = dict(('lexer_%d' % idx, item[0]) for idx,item in enumerate(lexicon))
        self.groupnames = groupnames

        # assemble regex parts to one regex
        igroupnames = dict((value,name) for name,value in six.iteritems(groupnames))

        regex_parts = ('(?P<%s>%s)' % (igroupnames[cls], regs) for cls,regs in lexicon)

        self.regex_string = '|'.join(regex_parts)
        self.regex = re.compile(self.regex_string)

    def lex(self, text):
        """
        Yield (token_type, data) tokens.
        The last token will be (EOF, None) where EOF
        """
        regex = self.regex
        groupnames = self.groupnames

        ignore = self.ignore or set()
        callbacks = self.callbacks or dict()

        position = 0
        size = len(text)

        while position < size:
            match = regex.match(text, position)
            if match is None:
                raise SVGError('Unknown token at position %d' %  position)

            position = match.end()
            cls = groupnames[match.lastgroup]
            value = match.group(match.lastgroup)

            if cls in ignore:
                continue

            if cls in callbacks:
                value = callbacks[cls](self, value)

            yield (cls, value)

        yield (EOF, None)

# Parse SVG length units
length_pattern = \
r"""
    ^                           # match start of line
    \s*                         # ignore whitespace
    (?P<value>[-\+]?\d*\.?\d*([eE][-\+]?\d+)?)  # match float or int value
    (?P<unit>.+)?               # match any chars
    \s*                         # ignore whitespace
    $                           # match end of line
"""

length_match = re.compile(length_pattern, re.X).match

def parseLength(length):
    """
    Convert length to reportlab points.
    """
    match = length_match(length)
    if match is None:
        raise SVGError("Not a valid length unit: '%s'" % length)

    value = match.group('value')
    unit = match.group('unit') or ''

    if not value:
        raise SVGError("Not a valid length unit: '%s'" % length)

    if not unit:
        if value[0] == 'e' or value[0] == 'E':
            return float('1' + value)
        else:
            return float(value)

    elif unit in ('em', 'ex', 'px', '%'):
        # ignoring relative units
        return float(value)

    elif unit == 'pc':
        return toLength(value + 'pica')

    elif unit in ('mm', 'cm', 'in', 'i', 'pt', 'pica'):
        return toLength(length)

    else:
        raise SVGError("Unknown unit '%s'" % unit)

# Parse SVG color definitions
color_pattern = \
r"""
^(
    (?P<named>
        \w+
    ) |

    (?P<hex>
        [#]
        (?P<hex_r>[a-zA-Z0-9]{2})
        (?P<hex_rg>[a-zA-Z0-9]{2})
        (?P<hex_rb>[a-zA-Z0-9]{2})
    ) |

    (?P<hexshort>
        [#]
        (?P<hexshort_r>[a-zA-Z0-9])
        (?P<hexshort_g>[a-zA-Z0-9])
        (?P<hexshort_b>[a-zA-Z0-9])
    ) |

    (?P<rgbint>
        [rR][gG][bB]
        \(
        (?P<rgbint_r>\d{1,3})
        ,\s?
        (?P<rgbint_g>\d{1,3})
        ,\s?
        (?P<rgbint_b>\d{1,3})
        \)
    ) |

    (?P<rgb>
        [rR][gG][bB]
        \(
        (?P<rgb_r>\d+\.?\d*)
        \%
        ,\s?
        (?P<rgb_g>\d+\.?\d*)
        \%
        ,\s?
        (?P<rgb_b>\d+\.?\d*)
        \%
        \)
    )
)$
"""

color_match = re.compile(color_pattern, re.X).match

NAMEDCOLOURS = {
    # expanded named colors from SVG spc
           'aliceblue' : colors.HexColor('#f0f8ff'),
        'antiquewhite' : colors.HexColor('#faebd7'),
                'aqua' : colors.HexColor('#00ffff'),
          'aquamarine' : colors.HexColor('#7fffd4'),
               'azure' : colors.HexColor('#f0ffff'),
               'beige' : colors.HexColor('#f5f5dc'),
              'bisque' : colors.HexColor('#ffe4c4'),
               'black' : colors.HexColor('#000000'),
      'blanchedalmond' : colors.HexColor('#ffebcd'),
                'blue' : colors.HexColor('#0000ff'),
          'blueviolet' : colors.HexColor('#8a2be2'),
               'brown' : colors.HexColor('#a52a2a'),
           'burlywood' : colors.HexColor('#deb887'),
           'cadetblue' : colors.HexColor('#5f9ea0'),
          'chartreuse' : colors.HexColor('#7fff00'),
           'chocolate' : colors.HexColor('#d2691e'),
               'coral' : colors.HexColor('#ff7f50'),
      'cornflowerblue' : colors.HexColor('#6495ed'),
            'cornsilk' : colors.HexColor('#fff8dc'),
             'crimson' : colors.HexColor('#dc143c'),
                'cyan' : colors.HexColor('#00ffff'),
            'darkblue' : colors.HexColor('#00008b'),
            'darkcyan' : colors.HexColor('#008b8b'),
       'darkgoldenrod' : colors.HexColor('#b8860b'),
            'darkgray' : colors.HexColor('#a9a9a9'),
           'darkgreen' : colors.HexColor('#006400'),
            'darkgrey' : colors.HexColor('#a9a9a9'),
           'darkkhaki' : colors.HexColor('#bdb76b'),
         'darkmagenta' : colors.HexColor('#8b008b'),
      'darkolivegreen' : colors.HexColor('#556b2f'),
          'darkorange' : colors.HexColor('#ff8c00'),
          'darkorchid' : colors.HexColor('#9932cc'),
             'darkred' : colors.HexColor('#8b0000'),
          'darksalmon' : colors.HexColor('#e9967a'),
        'darkseagreen' : colors.HexColor('#8fbc8b'),
       'darkslateblue' : colors.HexColor('#483d8b'),
       'darkslategray' : colors.HexColor('#2f4f4f'),
       'darkslategrey' : colors.HexColor('#2f4f4f'),
       'darkturquoise' : colors.HexColor('#00ced1'),
          'darkviolet' : colors.HexColor('#9400d3'),
            'deeppink' : colors.HexColor('#ff1493'),
         'deepskyblue' : colors.HexColor('#00bfff'),
             'dimgray' : colors.HexColor('#696969'),
             'dimgrey' : colors.HexColor('#696969'),
          'dodgerblue' : colors.HexColor('#1e90ff'),
           'firebrick' : colors.HexColor('#b22222'),
         'floralwhite' : colors.HexColor('#fffaf0'),
         'forestgreen' : colors.HexColor('#228b22'),
             'fuchsia' : colors.HexColor('#ff00ff'),
           'gainsboro' : colors.HexColor('#dcdcdc'),
          'ghostwhite' : colors.HexColor('#f8f8ff'),
                'gold' : colors.HexColor('#ffd700'),
           'goldenrod' : colors.HexColor('#daa520'),
                'gray' : colors.HexColor('#808080'),
               'green' : colors.HexColor('#008000'),
         'greenyellow' : colors.HexColor('#adff2f'),
                'grey' : colors.HexColor('#808080'),
            'honeydew' : colors.HexColor('#f0fff0'),
             'hotpink' : colors.HexColor('#ff69b4'),
           'indianred' : colors.HexColor('#cd5c5c'),
              'indigo' : colors.HexColor('#4b0082'),
               'ivory' : colors.HexColor('#fffff0'),
               'khaki' : colors.HexColor('#f0e68c'),
            'lavender' : colors.HexColor('#e6e6fa'),
       'lavenderblush' : colors.HexColor('#fff0f5'),
           'lawngreen' : colors.HexColor('#7cfc00'),
        'lemonchiffon' : colors.HexColor('#fffacd'),
           'lightblue' : colors.HexColor('#add8e6'),
          'lightcoral' : colors.HexColor('#f08080'),
           'lightcyan' : colors.HexColor('#e0ffff'),
'lightgoldenrodyellow' : colors.HexColor('#fafad2'),
           'lightgray' : colors.HexColor('#d3d3d3'),
          'lightgreen' : colors.HexColor('#90ee90'),
           'lightgrey' : colors.HexColor('#d3d3d3'),
           'lightpink' : colors.HexColor('#ffb6c1'),
         'lightsalmon' : colors.HexColor('#ffa07a'),
       'lightseagreen' : colors.HexColor('#20b2aa'),
        'lightskyblue' : colors.HexColor('#87cefa'),
      'lightslategray' : colors.HexColor('#778899'),
      'lightslategrey' : colors.HexColor('#778899'),
      'lightsteelblue' : colors.HexColor('#b0c4de'),
         'lightyellow' : colors.HexColor('#ffffe0'),
                'lime' : colors.HexColor('#00ff00'),
           'limegreen' : colors.HexColor('#32cd32'),
               'linen' : colors.HexColor('#faf0e6'),
             'magenta' : colors.HexColor('#ff00ff'),
              'maroon' : colors.HexColor('#800000'),
    'mediumaquamarine' : colors.HexColor('#66cdaa'),
          'mediumblue' : colors.HexColor('#0000cd'),
        'mediumorchid' : colors.HexColor('#ba55d3'),
        'mediumpurple' : colors.HexColor('#9370db'),
      'mediumseagreen' : colors.HexColor('#3cb371'),
     'mediumslateblue' : colors.HexColor('#7b68ee'),
   'mediumspringgreen' : colors.HexColor('#00fa9a'),
     'mediumturquoise' : colors.HexColor('#48d1cc'),
     'mediumvioletred' : colors.HexColor('#c71585'),
        'midnightblue' : colors.HexColor('#191970'),
           'mintcream' : colors.HexColor('#f5fffa'),
           'mistyrose' : colors.HexColor('#ffe4e1'),
            'moccasin' : colors.HexColor('#ffe4b5'),
         'navajowhite' : colors.HexColor('#ffdead'),
                'navy' : colors.HexColor('#000080'),
             'oldlace' : colors.HexColor('#fdf5e6'),
               'olive' : colors.HexColor('#808000'),
           'olivedrab' : colors.HexColor('#6b8e23'),
              'orange' : colors.HexColor('#ffa500'),
           'orangered' : colors.HexColor('#ff4500'),
              'orchid' : colors.HexColor('#da70d6'),
       'palegoldenrod' : colors.HexColor('#eee8aa'),
           'palegreen' : colors.HexColor('#98fb98'),
       'paleturquoise' : colors.HexColor('#afeeee'),
       'palevioletred' : colors.HexColor('#db7093'),
          'papayawhip' : colors.HexColor('#ffefd5'),
           'peachpuff' : colors.HexColor('#ffdab9'),
                'peru' : colors.HexColor('#cd853f'),
                'pink' : colors.HexColor('#ffc0cb'),
                'plum' : colors.HexColor('#dda0dd'),
          'powderblue' : colors.HexColor('#b0e0e6'),
              'purple' : colors.HexColor('#800080'),
                 'red' : colors.HexColor('#ff0000'),
           'rosybrown' : colors.HexColor('#bc8f8f'),
           'royalblue' : colors.HexColor('#4169e1'),
         'saddlebrown' : colors.HexColor('#8b4513'),
              'salmon' : colors.HexColor('#fa8072'),
          'sandybrown' : colors.HexColor('#f4a460'),
            'seagreen' : colors.HexColor('#2e8b57'),
            'seashell' : colors.HexColor('#fff5ee'),
              'sienna' : colors.HexColor('#a0522d'),
              'silver' : colors.HexColor('#c0c0c0'),
             'skyblue' : colors.HexColor('#87ceeb'),
           'slateblue' : colors.HexColor('#6a5acd'),
           'slategray' : colors.HexColor('#708090'),
           'slategrey' : colors.HexColor('#708090'),
                'snow' : colors.HexColor('#fffafa'),
         'springgreen' : colors.HexColor('#00ff7f'),
           'steelblue' : colors.HexColor('#4682b4'),
                 'tan' : colors.HexColor('#d2b48c'),
                'teal' : colors.HexColor('#008080'),
             'thistle' : colors.HexColor('#d8bfd8'),
              'tomato' : colors.HexColor('#ff6347'),
           'turquoise' : colors.HexColor('#40e0d0'),
              'violet' : colors.HexColor('#ee82ee'),
               'wheat' : colors.HexColor('#f5deb3'),
               'white' : colors.HexColor('#ffffff'),
          'whitesmoke' : colors.HexColor('#f5f5f5'),
              'yellow' : colors.HexColor('#ffff00'),
         'yellowgreen' : colors.HexColor('#9acd32'),
}

def parseColor(color):
    """
    Convert SVG color to reportlab color
    """
    if color == 'none':
        return None

    elif color == 'currentColor':
        return "currentColor"

    elif color == 'transparent':
        return colors.Color(0.,0.,0.,0.)

    match = color_match(color)
    if match is None:
        raise SVGError("Not a valid color definition: '%s'" % color)

    info = match.groupdict()

    if info['named'] is not None:
        if not color in NAMEDCOLOURS:
            raise SVGError("Not a valid named color: '%s'" % color)

        return NAMEDCOLOURS[color]

    elif info['hex'] is not None:
        return colors.HexColor(color)

    elif info['hexshort'] is not None:
        r = 2*info['hexshort_r']
        g = 2*info['hexshort_g']
        b = 2*info['hexshort_b']

        return colors.HexColor('#%s%s%s' % (r,g,b))

    elif info['rgbint'] is not None:
        r = int(info['rgbint_r'])
        g = int(info['rgbint_g'])
        b = int(info['rgbint_b'])

        if r > 255 or g > 255 or b > 255:
            raise SVGError("RGB value outside range: '%s'" % color)

        return colors.Color(r/255., g/255., b/255.)

    elif info['rgb'] is not None:
        r = float(info['rgb_r'])
        g = float(info['rgb_g'])
        b = float(info['rgb_b'])

        if r > 100 or g > 100 or b > 100:
            raise SVGError("RGB value outside range: '%s'" % color)

        return colors.Color(r/100., g/100., b/100.)

class SVGStyle(Lexer):
    """
    Break SVG inline style into tokens
    """
    name = object()
    value = object()
    delimiter = object()
    comment = object()

    lexicon = ( \
        (delimiter   , r'[ :;\n]'),
        (comment     , r'/\*.+\*/'),
        (name        , r'[\w\-#]+?(?=:)'),
        (value       , r'[\w\-#\.\(\)%,][\w \-#\.\(\)%,]*?(?=[;])'),
    )

    ignore = frozenset((delimiter,comment))

    def __init__(self):
        Lexer.__init__(self)

    def parse(self, text):
        """
        Parse a string of SVG <path> data.
        """
        if six.PY2:
            next = self.lex(text + ';').next
        else:
            next = self.lex(text + ';').__next__
        styles = {}

        while True:
            token, value = next()
            if token == EOF:
                break

            name = value.rstrip(':')

            token, value = next()

            if token == EOF:
                raise SVGError('expected value in style definition')

            if name in styles:
                raise SVGError('style redefined')

            styles[name] = value

        return styles


parseStyle = SVGStyle()

class SVGTransform(Lexer):
    """
    Break SVG transform into tokens.
    """
    numfloat = object()
    numint = object()
    string = object()
    skip = object()

    numbers = frozenset((numfloat, numint))

    lexicon = ( \
        (numfloat   , Lexer.Float),
        (numint     , Lexer.Int),
        (string     , r'\w+'),
        (skip       , r'[\(\), \n]'),
    )

    ignore = frozenset((skip,))

    callbacks = {
        numfloat    : lambda self,value: float(value),
        numint      : lambda self,value: float(value)
    }

    def __init__(self):
        Lexer.__init__(self)

    def assertion(self, condition, msg = ''):
        if not condition:
            raise SVGError(msg)

    def iterparse(self, text):
        """
        Parse a string of SVG transform data.
        """
        assertion = self.assertion
        if six.PY2:
            next = self.lex(text).next
        else:
            next = self.lex(text).__next__
        numbers = self.numbers
        string = self.string

        token, value = next()
        while token != EOF:
            assertion(token is string, 'Expected string')

            transform = value

            if transform == 'matrix':
                token, a = next()
                assertion(token in numbers, 'Expected number')

                token, b = next()
                assertion(token in numbers, 'Expected number')

                token, c = next()
                assertion(token in numbers, 'Expected number')

                token, d = next()
                assertion(token in numbers, 'Expected number')

                token, e = next()
                assertion(token in numbers, 'Expected number')

                token, f = next()
                assertion(token in numbers, 'Expected number')

                yield (transform, (a,b,c,d,e,f))

            elif transform == 'translate':
                token, tx = next()
                assertion(token in numbers, 'Expected number')

                token, value = next()
                ty = value
                if not token in numbers:
                    ty = 0.

                yield (transform, (tx, ty))

                if not token in numbers:
                    continue

            elif transform == 'scale':
                token, sx = next()
                assertion(token in numbers, 'Expected number')

                token, value = next()
                sy = value
                if not token in numbers:
                    sy = sx

                yield (transform, (sx, sy))

                if not token in numbers:
                    continue

            elif transform == 'rotate':
                token, angle = next()
                assertion(token in numbers, 'Expected number')

                token, value = next()
                cx = value

                if token in numbers:
                    token, value = next()
                    assertion(token in numbers, 'Expected number')

                    cy = value

                    yield (transform, (angle,(cx,cy)))

                else:
                    yield (transform, (angle,None))
                    continue

            elif transform == 'skewX' or transform == 'skewY':
                token, value = next()
                angle = value
                assertion(token in numbers, 'Expected number')

                yield (transform, (angle,))

            else:
                raise SVGError("unknown transform '%s'" % transform)

            # fetch next token
            token, value = next()

parseTransform = SVGTransform()

class SVGPath(Lexer):
    """
    Break SVG path data into tokens.

    The SVG spec requires that tokens are greedy.
    This lexer relies on Python's regexes defaulting to greediness.
    """
    numfloat = object()
    numint = object()
    numexp = object()
    string = object()
    skip = object

    lexicon = ( \
        (numfloat   , Lexer.Float),
        (numint     , Lexer.Int),
        (numexp     , r'(?:[Ee][-\+]?\d+)'),
        (string     , r'[AaCcHhLlMmQqSsTtVvZz]'),
        (skip       , r'[, \n]'),
    )

    ignore = frozenset((skip,))

    callbacks = {
        numfloat    : lambda self,value: float(value),
        numint      : lambda self,value: float(value),
        numexp      : lambda self,value: float('1.'+value)
    }

    numbers = frozenset((numfloat, numint, numexp))

    def __init__(self):
        Lexer.__init__(self)

    def assertion(self, condition, msg = ''):
        if not condition:
            raise SVGError(msg)

    def iterparse(self, text):
        """
        Parse a string of SVG <path> data.
        """
        assertion = self.assertion
        numbers = self.numbers
        string = self.string

        if six.PY2:
            next = self.lex(text).next
        else:
            next = self.lex(text).__next__

        token, value = next()
        while token != EOF:
            assertion(token is string, 'Expected string in path data')
            cmd = value
            CMD = value.upper()

            # closePath
            if CMD in 'Z':
                token, value = next()
                yield (cmd, (None,))

            # moveTo, lineTo, curve, smoothQuadraticBezier, quadraticBezier, smoothCurve
            elif CMD in 'CMLTQS':
                coords = []
                token, value = next()
                while token in numbers:
                    last = value

                    token, value = next()
                    assertion(token in numbers, 'Expected number in path data')

                    coords.append((last,value))

                    token, value = next()

                if CMD == 'C':
                    assertion(len(coords) % 3 == 0, 'Expected coordinate triplets in path data')

                yield (cmd, tuple(coords))

            # horizontalLine or verticalLine
            elif CMD in 'HVhv':
                coords = []
                token, value = next()
                assertion(token in numbers, 'Expected number')

                while token in numbers:
                    coords.append(value)
                    token, value = next()

                yield (cmd, tuple(coords))

            # ellipticalArc
            elif CMD == 'A':
                token, rx = next()
                #assertion(token in numbers and value > 0, 'expected positive number in path data')
                rx = value

                token, ry = next()
                #assertion(token in numbers and ry > 0, 'expected positive number in path data')

                token, rotation = next()
                #assertion(token in numbers, 'expected number in path data')

                token, largearc = next()
                #assertion(token is int and largearc in (0,1), 'expected 0 or 1 in path data')

                token, sweeparc = next()
                #assertion(token is int and sweeparc in (0,1), 'expected 0 or 1 in path data')

                token, x = next()
                #assertion(token in numbers, 'expected number in path data')

                token, y = next()
                #assertion(token in numbers, 'expected number in path data')

                yield (cmd, ((rx,ry), rotation, largearc, sweeparc, (x,y)))

                token, value = next()

            else:
                raise SVGError("cmd '%s' in path data not supported" % cmd)

parsePath = SVGPath()

def parseDashArray(array):
    if array == 'none':
        return None

    return map(parseLength, re.split('[ ,]+', array))

def parseOpacity(value):
    try:
        opacity = float(value)
    except ValueError:
        raise SVGError('expected float value')

    # clamp value
    opacity = min(max(opacity, 0.), 1.)

    return opacity

def parseAnchor(value):
    if not value or value == 'none':
        return 'start'

    if value not in ('start', 'middle', 'end'):
        raise SVGError("unknown '%s' alignment" % value)

    return value

# map svg font names to reportlab names
FONTMAPPING = {
    "sans-serif"    : "Helvetica",
    "serif"         : "Times-Roman",
    "monospace"     : "Courier"
}

# map svg style to repotlab attributes and converters
STYLES = {
    "fill"              : ("fillColor",         parseColor),
    "stroke"            : ("strokeColor",       parseColor),
    "stroke-width"      : ("strokeWidth",       parseLength),
    "stroke-linejoin"   : ("strokeLineJoin",    lambda style: {"miter":0, "round":1, "bevel":2}[style]),
    "stroke-linecap"    : ("strokeLineCap",     lambda style: {"butt":0, "round":1, "square":2}[style]),
    "stroke-dasharray"  : ("strokeDashArray",   parseDashArray),
    "fill-opacity"      : ('fillOpacity',       parseOpacity),
    "stroke-opacity"    : ('strokeOpacity',     parseOpacity),
    "font-family"       : ("fontName",          lambda name: FONTMAPPING.get(name, 'Helvetica')),
    "font-size"         : ("fontSize",          parseLength),
    "text-anchor"       : ("textAnchor",        parseAnchor),
}

STYLE_NAMES = frozenset(list(STYLES.keys()) + ['color',])
STYLES_FONT = frozenset(('font-family','font-size','text-anchor'))

class Renderer:
    LINK = '{http://www.w3.org/1999/xlink}href'

    SVG_DEFS        = '{http://www.w3.org/2000/svg}defs'
    SVG_ROOT        = '{http://www.w3.org/2000/svg}svg'
    SVG_A           = '{http://www.w3.org/2000/svg}a'
    SVG_G           = '{http://www.w3.org/2000/svg}g'
    SVG_SYMBOL      = '{http://www.w3.org/2000/svg}symbol'
    SVG_USE         = '{http://www.w3.org/2000/svg}use'
    SVG_RECT        = '{http://www.w3.org/2000/svg}rect'
    SVG_CIRCLE      = '{http://www.w3.org/2000/svg}circle'
    SVG_ELLIPSE     = '{http://www.w3.org/2000/svg}ellipse'
    SVG_LINE        = '{http://www.w3.org/2000/svg}line'
    SVG_POLYLINE    = '{http://www.w3.org/2000/svg}polyline'
    SVG_POLYGON     = '{http://www.w3.org/2000/svg}polygon'
    SVG_PATH        = '{http://www.w3.org/2000/svg}path'
    SVG_TEXT        = '{http://www.w3.org/2000/svg}text'
    SVG_TSPAN       = '{http://www.w3.org/2000/svg}tspan'
    SVG_IMAGE       = '{http://www.w3.org/2000/svg}image'

    SVG_NODES = frozenset((SVG_ROOT, SVG_A, SVG_G, SVG_SYMBOL, SVG_USE, SVG_RECT,
                           SVG_CIRCLE, SVG_ELLIPSE, SVG_LINE, SVG_POLYLINE,
                           SVG_POLYGON, SVG_PATH, SVG_TEXT, SVG_IMAGE))

    def __init__(self, filename):
        self.filename = filename
        self.level = 0
        self.styles = {}
        self.mainGroup = Group()
        self.drawing = None
        self.root = None

    def render(self, node, parent = None):
        if parent == None:
            parent = self.mainGroup

        # ignore if display = none
        display = node.get('display')
        if display == "none":
            return

        if node.tag == self.SVG_ROOT:
            self.level += 1

            if not self.drawing is None:
                raise SVGError('drawing already created!')

            self.root = node

            # default styles
            style = {
                'color':'black',
                'fill':'black',
                'stroke':'none',
                'font-family':'Helvetica',
                'font-size':'12'
            }

            self.styles[self.level] = style

            # iterate children
            for child in node:
                self.render(child, self.mainGroup)

            # create drawing
            width = node.get('width', '100%')
            height = node.get('height', '100%')

            if node.get("viewBox"):
                try:
                    minx, miny, width, height = node.get("viewBox").split()
                except ValueError:
                    raise SVGError("viewBox values not valid")

            if width.endswith('%') and height.endswith('%'):
                # handle relative size
                wscale = parseLength(width) / 100.
                hscale = parseLength(height) / 100.

                xL,yL,xH,yH =  self.mainGroup.getBounds()
                self.drawing = Drawing(xH*wscale + xL, yH*hscale + yL)

            else:
                self.drawing = Drawing(parseLength(width), parseLength(height))

            height = self.drawing.height
            self.mainGroup.scale(1, -1)
            self.mainGroup.translate(-float(minx), -height-float(miny))
            self.drawing.add(self.mainGroup)

            self.level -= 1

            return self.drawing

        elif node.tag in (self.SVG_G, self.SVG_A, self.SVG_SYMBOL):
            self.level += 1

            # set this levels style
            style = self.styles[self.level - 1].copy()
            style = self.nodeStyle(node, style)
            self.styles[self.level] = style

            group = Group()

            # iterate children
            for child in node:
                self.render(child, group)

            parent.add(group)

            transforms = node.get('transform')
            if transforms:
                for op in parseTransform.iterparse(transforms):
                    self.applyTransformOnGroup(group, op)

            self.level -= 1

        elif node.tag == self.SVG_USE:
            self.level += 1

            # set this levels style
            style = self.styles[self.level - 1].copy()
            style = self.nodeStyle(node, style)
            self.styles[self.level] = style

            group = Group()

            # link id
            link_id = node.get(self.LINK).lstrip('#')

            # find linked node in defs section
            target = None
            for defs in self.root.getiterator(self.SVG_DEFS):
                for element in defs:
                    if element.get('id') == link_id:
                        target = element
                        break

            if target is None:
                raise SVGError("Could not find use node '%s'" % link_id)

            self.render(target, group)

            parent.add(group)

            # apply transform
            transforms = node.get('transform')
            if transforms:
                for op in parseTransform.iterparse(transforms):
                    self.applyTransformOnGroup(group, op)

            # apply 'x' and 'y' attribute as translation of defs object
            if node.get('x') or node.get('y'):
                dx = parseLength(node.get('x','0'))
                dy = parseLength(node.get('y','0'))

                self.applyTransformOnGroup(group, ('translate', (dx,dy)))

            self.level -= 1

        elif node.tag == self.SVG_LINE:
            # get coordinates
            x1 = parseLength(node.get('x1', '0'))
            y1 = parseLength(node.get('y1', '0'))
            x2 = parseLength(node.get('x2', '0'))
            y2 = parseLength(node.get('y2', '0'))

            shape = Line(x1, y1, x2, y2)
            self.addShape(parent, node, shape)

        elif node.tag == self.SVG_RECT:
            # get coordinates
            x = parseLength(node.get('x', '0'))
            y = parseLength(node.get('y', '0'))
            width = parseLength(node.get('width'))
            height = parseLength(node.get('height'))

            rx = parseLength(node.get('rx', '0'))
            ry = parseLength(node.get('ry', '0'))

            shape = Rect(x, y, width, height, rx=rx, ry=ry)
            self.addShape(parent, node, shape)

        elif node.tag == self.SVG_CIRCLE:
            cx = parseLength(node.get('cx', '0'))
            cy = parseLength(node.get('cy', '0'))
            r = parseLength(node.get('r'))

            if r > 0.:
                shape = Circle(cx, cy, r)
                self.addShape(parent, node, shape)

        elif node.tag == self.SVG_ELLIPSE:
            cx = parseLength(node.get('cx', '0'))
            cy = parseLength(node.get('cy', '0'))
            rx = parseLength(node.get('rx'))
            ry = parseLength(node.get('ry'))

            if rx > 0. and ry > 0.:
                shape = Ellipse(cx, cy, rx, ry)
                self.addShape(parent, node, shape)

        elif node.tag == self.SVG_POLYLINE:
            # convert points
            points = node.get('points').strip()
            if len(points) == 0:
                return

            points = map(parseLength, re.split('[ ,]+', points))

            # Need to use two shapes, because standard RLG polylines
            # do not support filling...
            group = Group()
            shape = Polygon(points)
            self.applyStyleToShape(shape, node)
            shape.strokeColor = None
            group.add(shape)

            shape = PolyLine(points)
            self.applyStyleToShape(shape, node)
            group.add(shape)

            self.addShape(parent, node, group)

        elif node.tag == self.SVG_POLYGON:
            # convert points
            points = node.get('points').strip()
            if len(points) == 0:
                return

            points = map(parseLength, re.split('[ ,]+', points))

            shape = Polygon(points)
            self.addShape(parent, node, shape)

        elif node.tag == self.SVG_IMAGE:
            x = parseLength(node.get('x', '0'))
            y = parseLength(node.get('y', '0'))
            width = parseLength(node.get('width', '0'))
            height = parseLength(node.get('height', '0'))

            # link id
            link_id = node.get(self.LINK)

            filename = os.path.join(os.path.dirname(self.filename), link_id)
            shape = Image(x, y, width, height, filename)

            self.addShape(parent, node, shape)

        elif node.tag == self.SVG_TEXT:
            # Todo:
            # - rotation not handled
            # - baseshift not handled
            # - embedded span node not handled
            #
            def parsePos(node, subnode, name, default = '0'):
                values = subnode.get(name)
                if values is None:
                    if not node is None:
                        values = node.get(name, default)
                    else:
                        values = default

                return map(parseLength, values.split())

            def getPos(values, i, default = None):
                if i >= len(values):
                    if default is None:
                        return values[-1]
                    else:
                        return default
                else:
                    return values[i]

            def handleText(node, subnode, text):
                # get position variables
                xs = parsePos(node, subnode, 'x')
                dxs = parsePos(node, subnode, 'dx')
                ys = parsePos(node, subnode,'y')
                dys = parsePos(node, subnode,'dy')

                if sum(map(len, (xs,ys,dxs,dys))) == 4:
                    # single value
                    shape = String(xs[0] + dxs[0], -ys[0] - dys[0], text)
                    self.applyStyleToShape(shape, subnode)
                    group.add(shape)

                else:
                    # multiple values
                    for i,c in enumerate(text):
                        x = getPos(xs, i)
                        dx = getPos(dxs, i, 0)
                        y = getPos(ys, i)
                        dy = getPos(dys, i, 0)

                        shape = String(x + dx, -y -dy, c)
                        self.applyStyleToShape(shape, subnode)
                        group.add(shape)

            if node.text and node.text.strip():
                group = Group()

                handleText(None, node, node.text.strip())

                group.scale(1, -1)

                self.addShape(parent, node, group)

            if len(node) > 0:
                group = Group()

                self.level += 1

                # set this levels style
                style = self.styles[self.level - 1].copy()
                nodestylestyle = self.nodeStyle(node, style)
                self.styles[self.level] = nodestylestyle

                for subnode in node:
                    if subnode.tag == self.SVG_TSPAN:
                        handleText(node, subnode, subnode.text.strip())

                self.level -= 1

                group.scale(1, -1)
                self.addShape(parent, node, group)


        elif node.tag == self.SVG_PATH:
            def convertQuadratic(Q0, Q1, Q2):
                C1 = (Q0[0] + 2./3*(Q1[0] - Q0[0]), Q0[1] + 2./3*(Q1[1] - Q0[1]))
                C2 = (C1[0] + 1./3*(Q2[0] - Q0[0]), C1[1] + 1./3*(Q2[1] - Q0[1]))
                C3 = Q2
                return C1[0], C1[1], C2[0], C2[1], C3[0], C3[1]

            def prevCtrl(lastOp, lastArgs, currentX, currentY):
                # fetch last controll point
                if lastOp in 'CScsQqTt':
                    x, y = lastArgs[-2]

                    # mirror about current point
                    return currentX + (currentX-x), currentY + (currentY-y)

                else:
                    # defaults to current point
                    return currentX, currentY

            # store sub paths in 'paths' list
            shape = Path()

            # keep track of current point and path start point
            startX, startY = 0.,0.
            currentX, currentY = 0.,0.

            # keep track of last operation
            lastOp = None
            lastArgs = None

            # avoid empty path data
            data = node.get('d')
            if data is None or len(data) == 0:
                return

            for op, args in parsePath.iterparse(data):
                if op == 'z' or op == 'Z':
                    # close path or subpath
                    shape.closePath()

                    # next sub path starts at begining of current path
                    currentX, currentY = startX, startY

                elif op == 'M':
                    # moveto absolute
                    if not lastOp is None and lastOp not in ('z', 'Z'):
                        # close sub path
                        shape.closePath()

                    x, y = args[0]
                    shape.moveTo(x, y)

                    startX, startY = x, y

                    # multiple moveto arge result in line
                    for x, y in args[1:]:
                        shape.lineTo(x, y)

                    currentX, currentY = x, y

                elif op == 'm':
                    if not lastOp is None and lastOp not in ('z', 'Z'):
                        # close sub path
                        shape.closePath()

                    # moveto relative
                    rx, ry = args[0]
                    x, y = currentX + rx, currentY + ry
                    shape.moveTo(x, y)

                    startX, startY = x, y
                    currentX, currentY = x, y

                    # multiple moveto arge result in line
                    for rx, ry in args[1:]:
                        x, y = currentX + rx, currentY + ry
                        shape.lineTo(x, y)
                        currentX, currentY = x, y

                elif op == 'L':
                    # lineto absolute
                    for x,y in args:
                        shape.lineTo(x, y)

                    currentX, currentY = x, y

                elif op == 'l':
                    # lineto relative
                    for rx,ry in args:
                        x, y = currentX + rx, currentY + ry
                        shape.lineTo(x, y)
                        currentX, currentY = x, y

                elif op == 'V':
                    # vertical line absolute
                    for y in args:
                        shape.lineTo(currentX, y)

                    currentY = y

                elif op == 'v':
                    # vertical line relative
                    for ry in args:
                        y = currentY + ry
                        shape.lineTo(currentX, y)
                        currentY = y

                elif op == 'H':
                    # horisontal line absolute
                    for x in args:
                        shape.lineTo(x, currentY)

                    currentX = x

                elif op == 'h':
                    # horisontal line relative
                    for rx in args:
                        x = currentX + rx
                        shape.lineTo(x, currentY)

                    currentX = x

                elif op == 'C':
                    # cubic bezier absolute
                    for i in range(len(args)//3):
                        x1, y1 = args[i*3 + 0]
                        x2, y2 = args[i*3 + 1]
                        x3, y3 = args[i*3 + 2]

                        shape.curveTo(x1, y1, x2, y2, x3, y3)

                    currentX, currentY = x3, y3

                elif op == 'c':
                    # cubic bezier relative
                    for i in range(len(args)//3):
                        x1, y1 = args[i*3 + 0]
                        x1, y1 = x1 + currentX, y1 + currentY

                        x2, y2 = args[i*3 + 1]
                        x2, y2 = x2 + currentX, y2 + currentY

                        x3, y3 = args[i*3 + 2]
                        x3, y3 = x3 + currentX, y3 + currentY

                        shape.curveTo(x1, y1, x2, y2, x3, y3)

                        currentX, currentY = x3, y3

                    lastOp = op
                    lastArgs = (x1, y1), (x2, y2), (x3, y3)
                    continue

                elif op == 'S':
                    # shorthand cubic bezier absolute
                    for i in range(len(args)//2):
                        x1, y1 = prevCtrl(lastOp, lastArgs, currentX, currentY)

                        x2, y2 = args[i*2 + 0]
                        x3, y3 = args[i*2 + 1]

                        shape.curveTo(x1, y1, x2, y2, x3, y3)

                        lastOp = op
                        lastArgs = (x2, y2), (x3, y3)

                        currentX, currentY = x3, y3

                    continue

                elif op == 's':
                    # shorthand cubic bezier relative
                    for i in range(len(args)//2):
                        x1, y1 = prevCtrl(lastOp, lastArgs, currentX, currentY)

                        x2, y2 = args[i*2 + 0]
                        x2, y2 = x2 + currentX, y2 + currentY

                        x3, y3 = args[i*2 + 1]
                        x3, y3 = x3 + currentX, y3 + currentY

                        shape.curveTo(x1, y1, x2, y2, x3, y3)

                        currentX, currentY = x3, y3

                        lastOp = op
                        lastArgs = (x1, y1), (x2, y2), (x3, y3)

                    continue

                elif op == 'Q':
                    # quadratic bezier absolute
                    for i in range(len(args)//2):
                        x1, y1 = currentX, currentY
                        x2, y2 = args[i*2 + 0]
                        x3, y3 = args[i*2 + 1]

                        ctrls = convertQuadratic((x1, y1), (x2, y2), (x3, y3))

                        shape.curveTo(*ctrls)

                        currentX, currentY = x3, y3

                    lastOp = op
                    lastArgs = (x2, y2), (x3, y3)
                    continue

                elif op == 'q':
                    # quadratic bezier relative
                    for i in range(len(args)//2):
                        x1, y1 = currentX, currentY

                        x2, y2 = args[i*2 + 0]
                        x2, y2 = x2 + currentX, y2 + currentY

                        x3, y3 = args[i*2 + 1]
                        x3, y3 = x3 + currentX, y3 + currentY

                        ctrls = convertQuadratic((x1, y1), (x2, y2), (x3, y3))

                        shape.curveTo(*ctrls)
                        currentX, currentY = x3, y3

                    lastOp = op
                    lastArgs = (x2, y2), (x3, y3)
                    continue

                elif op == 'T':
                    # shorthand quadratic bezier absolute
                    for i in range(len(args)):
                        x1, y1 = currentX, currentY
                        x2, y2 = prevCtrl(lastOp, lastArgs, currentX, currentY)
                        x3, y3 = args[i]

                        ctrls = convertQuadratic((x1, y1), (x2, y2), (x3, y3))
                        shape.curveTo(*ctrls)

                        currentX, currentY = x3, y3

                        lastOp = op
                        lastArgs = (x2, y2), (x3, y3)

                    continue


                elif op == 't':
                    # shorthand quadratic bezier relative
                    for i in range(len(args)):
                        x1, y1 = currentX, currentY
                        x2, y2 = prevCtrl(lastOp, lastArgs, currentX, currentY)

                        x3, y3 = args[i]
                        x3, y3 = x3 + currentX, y3 + currentY

                        ctrls = convertQuadratic((x1, y1), (x2, y2), (x3, y3))
                        shape.curveTo(*ctrls)

                        currentX, currentY = x3, y3

                        lastOp = op
                        lastArgs = (x2, y2), (x3, y3)

                    continue

                elif op == 'A' or op == 'a':
                    # elliptic arc missing
                    continue

                lastOp = op
                lastArgs = args

            # check if fill applies to path
            fill = None
            if node.get('fill'):
                # inline style
                fill = node.get('fill')
            else:
                # try local style
                if node.get('style'):
                    style = parseStyle.parse(node.get('style'))

                    if 'fill' in style:
                        fill = style['fill']

                # try global style
                if fill is None:
                    style = self.styles[self.level]

                    if 'fill' in style:
                        fill = style['fill']
                    else:
                        fill = 'none'

            # hack because RLG has no "semi-closed" paths...
            if lastOp == 'z' or lastOp == 'Z' or fill == 'none':
                self.addShape(parent, node, shape)

            else:
                group = Group()

                strokeshape = shape.copy()
                self.addShape(group, node, strokeshape, fill = 'none')

                shape.closePath()
                self.addShape(group, node, shape, stroke = 'none')

                self.addShape(parent, node, group)


    def addShape(self, parent, node, shape, **kwargs):

        self.applyStyleToShape(shape, node, **kwargs)

        transform = node.get('transform')
        if transform:
            # transform can only be applied to a group
            if isinstance(node, Group):
                group = node
            else:
                group = Group()

            for op in parseTransform.iterparse(transform):
                self.applyTransformOnGroup(group, op)

            if not isinstance(node, Group):
                group.add(shape)

            parent.add(group)

        else:
            parent.add(shape)

        return shape

    def nodeStyle(self, node, style = None):
        if style is None:
            style = {}

        # update with local style
        if node.get('style'):
            localstyle = parseStyle.parse(node.get('style'))
            for name, value in six.iteritems(localstyle):
                style[name] = value

        # update with inline style
        for name in STYLE_NAMES:
            value = node.get(name)
            if value:
                style[name] = value

        return style

    def applyStyleToShape(self, shape, node, **kwargs):
        # fetch global styles for this level
        globalstyle = self.styles[self.level]

        # update with node style
        localstyle = self.nodeStyle(node)

        for name in STYLES:
            # avoid text styles if not a String subclass
            if not isinstance(shape, String) and name in STYLES_FONT:
                continue

            value = None
            if name in kwargs:
                value = kwargs[name]
            elif name in localstyle:
                value = localstyle[name]
            elif name in globalstyle:
                value = globalstyle[name]
            else:
                continue

            # handle 'currentColor'
            if value == 'currentColor':
                if 'color' in localstyle:
                    value = localstyle['color']
                elif 'color' in globalstyle:
                    value = globalstyle['color']

            # fetch shape attribute name and converter
            attrname, converter = STYLES[name]

            # Sett attribute of shape
            if hasattr(shape, attrname):
                setattr(shape, attrname, converter(value))

        # defaults
        if isinstance(shape, String):
            if shape.fillColor is None:
                shape.fillColor = colors.black

        elif isinstance(shape, (Line, PolyLine)):
            if shape.strokeColor is None:
                shape.strokeColor = colors.black

        elif isinstance(shape, (Rect, Circle, Ellipse, Polygon)):
            if shape.fillColor is None and shape.strokeColor is None:
                shape.strokeColor = colors.black

    def applyTransformOnGroup(self, group, transform):
        op, args = transform

        if op == 'scale':
            group.scale(*args)

        elif op == 'translate':
            group.translate(*args)

        elif op == 'rotate':
            if len(args) == 2:
                angle, center = args
                if center:
                    cx, cy = center
                    group.translate(cx, cy)
                    group.rotate(angle)
                    group.translate(-cx, -cy)
                else:
                    group.rotate(angle)

        elif op == "skewX":
            angle = args[0]
            group.skew(angle, 0)

        elif op == "skewY":
            angle = args[0]
            group.skew(0, angle)

        elif op == "matrix":
            group.transform = mmult(group.transform, args)

def readFile(filename):
    """
    Open svg file and return root xml object
    """
    GZIPMAGIC = '\037\213'

    try:
        fh = open(filename, 'rb')
    except IOError:
        raise SVGError("could not open file '%s' for reading" % filename)

    # test for gzip compression
    magic = fh.read(2)
    fh.close()

    if magic == GZIPMAGIC:
        svg = cElementTree.parse(GzipFile(filename, 'r'))
    else:
        svg = cElementTree.parse(filename)

    root = svg.getroot()

    if not root.tag == Renderer.SVG_ROOT:
        raise SVGError("Expected SVG fragment as root object")

    return root

def svg2rlg(filename):
    """
    Open svg file and return reportlab drawing object
    """
    xml = readFile(filename)
    renderer = Renderer(filename)

    return renderer.render(xml)

if __name__ == "__main__":
    import sys
    import os

    #sys.argv.append('compliance/paths-data-15-t.svg')

    source = sys.argv[1]
    path, filename = os.path.split(source)
    name, ext = os.path.splitext(filename)

    drawing = svg2rlg(source)
    drawing.save(formats=['pdf'],outDir='.',fnRoot=name)

    os.startfile(name + '.pdf')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Template Support
"""
import zope
from z3c.rml import rml2pdf

try:
    import zope.pagetemplate.pagetemplatefile
except ImportError as err:
    raise
    ## zope.pagetemplate package has not been installed, uncomment this to mock
    # import types
    # zope.pagetemplate = types.ModuleType('pagetemplate')
    # zope.pagetemplate.pagetemplatefile = types.ModuleType('pagetemplatefile')
    # zope.pagetemplate.pagetemplatefile.PageTemplateFile = object


class RMLPageTemplateFile(zope.pagetemplate.pagetemplatefile.PageTemplateFile):

    def pt_getContext(self, args=(), options=None, **ignore):
        rval = {'template': self,
                'args': args,
                'nothing': None,
                'context': options
                }
        rval.update(self.pt_getEngine().getBaseNames())
        return rval

    def __call__(self, *args, **kwargs):
        rml = super(RMLPageTemplateFile, self).__call__(*args, **kwargs)

        return rml2pdf.parseString(
            rml, filename=self.pt_source_file()).getvalue()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing edge cases.
"""
import logging
import os
import unittest
import sys
import z3c.rml.tests
from PyPDF2.utils import PdfReadError
from z3c.rml import rml2pdf
from z3c.rml import pdfinclude

HERE = os.path.dirname(z3c.rml.tests.__file__)
LOG_FILE = os.path.join(os.path.dirname(__file__), 'render.log')


class RMLEdgeTests(unittest.TestCase):

    def setUp(self):
        import z3c.rml.tests.module
        sys.modules['module'] = z3c.rml.tests.module
        sys.modules['mymodule'] = z3c.rml.tests.module

    def tearDown(self):
        del sys.modules['module']
        del sys.modules['mymodule']

        from z3c.rml.document import LOGGER_NAME
        for handler in logging.getLogger(LOGGER_NAME).handlers:
            if handler.baseFilename == LOG_FILE:
                handler.close()

        if os.path.exists(LOG_FILE):
            os.remove(LOG_FILE)

    def test_duplicate_key(self):
        save_strict = pdfinclude.STRICT
        inpath = os.path.join(HERE, 'input', 'data',
                              'tag-includePdfPages-edge-dupekey.rml')
        outpath = os.path.join(HERE, 'output',
                               'tag-includePdfPages-edge-dupekey.pdf')

        # in strict mode include will fail because the PDF is bad
        pdfinclude.STRICT = True
        with self.assertRaises(PdfReadError):
            rml2pdf.go(inpath, outpath)

        # in non strict mode such errors are demoted to warnings
        pdfinclude.STRICT = False
        rml2pdf.go(inpath, outpath)

        pdfinclude.STRICT = save_strict


def test_suite():
    suite = unittest.TestSuite([unittest.makeSuite(RMLEdgeTests)])
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML Reference Generator
"""
import copy
import re
import os
import pygments.token
import zope.schema
import zope.schema.interfaces
from lxml import etree
from xml.sax import saxutils
from pygments.lexers import XmlLexer
from z3c.rml import attr, document, interfaces, pagetemplate, __version__


INPUT_URL = ('https://github.com/zopefoundation/z3c.rml/blob/master/src/z3c/'
             'rml/tests/input/%s')
EXPECTED_URL = ('https://github.com/zopefoundation/z3c.rml/blob/master/src/z3c/'
                'rml/tests/expected/%s?raw=true')

EXAMPLES_DIRECTORY = os.path.join(os.path.dirname(__file__), 'tests', 'input')
IGNORE_ATTRIBUTES = ('RMLAttribute', 'BaseChoice')
CONTENT_FIELD_TYPES = (
    attr.TextNode, attr.TextNodeSequence, attr.TextNodeGrid,
    attr.RawXMLContent, attr.XMLContent)
STYLES_FORMATTING = {
     pygments.token.Name.Tag : ('<font textColor="red">', '</font>'),
     pygments.token.Literal.String : ('<font textColor="blue">', '</font>'),
    }
EXAMPLE_NS = 'http://namespaces.zope.org/rml/doc'
EXAMPLE_ATTR_NAME = '{%s}example' %EXAMPLE_NS


def dedent(rml):
    spaces = re.findall('\n( *)<', rml)
    if not spaces:
        return rml
    least = min([len(s) for s in spaces if s != ''])
    return rml.replace('\n'+' '*least, '\n')


def enforceColumns(rml, columns=80):
    result = []
    for line in rml.split('\n'):
        if len(line) <= columns:
            result.append(line)
            continue
        # Determine the indentation for all other lines
        lineStart = re.findall('^( *<[a-zA-Z0-9]+ )', line)
        lineIndent = 0
        if lineStart:
            lineIndent = len(lineStart[0])
        # Create lines having at most the specified number of columns
        while len(line) > columns:
            end = line[:columns].rfind(' ')
            result.append(line[:end])
            line = ' '*lineIndent + line[end+1:]
        result.append(line)

    return '\n'.join(result)

def highlightRML(rml):
    lexer = XmlLexer()
    styledRml = ''
    for ttype, token in lexer.get_tokens(rml):
        start, end = STYLES_FORMATTING.get(ttype, ('', ''))
        styledRml += start + saxutils.escape(token) + end
    return styledRml


def removeDocAttributes(elem):
    for name in elem.attrib.keys():
        if name.startswith('{'+EXAMPLE_NS+'}'):
            del elem.attrib[name]
    for child in elem.getchildren():
        removeDocAttributes(child)


def getAttributeTypes():
    types = []
    candidates = sorted(attr.__dict__.items(), key=lambda e: e[0])
    for name, candidate in candidates:
        if not (isinstance(candidate, type) and
                zope.schema.interfaces.IField.implementedBy(candidate) and
                name not in IGNORE_ATTRIBUTES):
            continue
        types.append({
            'name': name,
            'description': candidate.__doc__
            })
    return types


def formatField(field):
    return field.__class__.__name__

def formatChoice(field):
    choices = ', '.join([repr(choice) for choice in field.choices.keys()])
    return '%s of (%s)' %(field.__class__.__name__, choices)

def formatSequence(field):
    vtFormatter = typeFormatters.get(field.value_type.__class__, formatField)
    return '%s of %s' %(field.__class__.__name__, vtFormatter(field.value_type))

def formatGrid(field):
    vtFormatter = typeFormatters.get(field.value_type.__class__, formatField)
    return '%s with %i cols of %s' %(
        field.__class__.__name__, field.columns, vtFormatter(field.value_type))

def formatCombination(field):
    vts = [typeFormatters.get(vt.__class__, formatField)(vt)
           for vt in field.value_types]
    return '%s of %s' %(field.__class__.__name__, ', '.join(vts))

typeFormatters = {
    attr.Choice: formatChoice,
    attr.Sequence: formatSequence,
    attr.Combination: formatCombination,
    attr.TextNodeSequence: formatSequence,
    attr.TextNodeGrid: formatGrid}

def processSignature(name, signature, examples, directives=None):
    if directives is None:
        directives = {}
    # Process this directive
    if signature not in directives:
        info = {'name': name, 'description': signature.getDoc(),
                'id': str(hash(signature)), 'deprecated': False}
        # If directive is deprecated, then add some info
        if interfaces.IDeprecatedDirective.providedBy(signature):
            info['deprecated'] = True
            info['reason'] = signature.getTaggedValue('deprecatedReason')
        attrs = []
        content = None
        for fname, field in zope.schema.getFieldsInOrder(signature):
            # Handle the case, where the field describes the content
            typeFormatter = typeFormatters.get(field.__class__, formatField)
            fieldInfo = {
                'name': fname,
                'type': typeFormatter(field),
                'title': field.title,
                'description': field.description,
                'required': field.required,
                'deprecated': False,
                }
            if field.__class__ in CONTENT_FIELD_TYPES:
                content = fieldInfo
            else:
                attrs.append(fieldInfo)

            # Add a separate entry for the deprecated field
            if interfaces.IDeprecated.providedBy(field):
                deprFieldInfo = fieldInfo.copy()
                deprFieldInfo['deprecated'] = True
                deprFieldInfo['name'] = field.deprecatedName
                deprFieldInfo['reason'] = field.deprecatedReason
                attrs.append(deprFieldInfo)

        info['attributes'] = attrs
        info['content'] = content
        # Examples can be either gotten by interface path or tag name
        ifacePath = signature.__module__ + '.' + signature.__name__
        if ifacePath in examples:
            info['examples'] = examples[ifacePath]
        else:
            info['examples'] = examples.get(name, None)

        subs = []
        for occurence in signature.queryTaggedValue('directives', ()):
            subs.append({
                'name': occurence.tag,
                'occurence': occurence.__class__.__name__,
                'deprecated': interfaces.IDeprecatedDirective.providedBy(
                                 occurence.signature),
                'id': str(hash(occurence.signature))
                })
        info['sub-directives'] = subs
        directives[signature] = info
    # Process Children
    for occurence in signature.queryTaggedValue('directives', ()):
        processSignature(occurence.tag, occurence.signature,
                         examples, directives)


def extractExamples(directory):
    examples = {}
    for filename in os.listdir(directory):
        if not filename.endswith('.rml'):
            continue
        rmlFile = open(os.path.join(directory, filename), 'rb')
        root = etree.parse(rmlFile).getroot()
        elements = root.xpath('//@doc:example/parent::*',
                              namespaces={'doc': EXAMPLE_NS})
        # Phase 1: Collect all elements
        for elem in elements:
            demoTag = elem.get(EXAMPLE_ATTR_NAME) or elem.tag
            elemExamples = examples.setdefault(demoTag, [])
            elemExamples.append({
                'filename': filename,
                'line': elem.sourceline,
                'element': elem,
                'rmlurl': INPUT_URL %filename,
                'pdfurl': EXPECTED_URL %(filename[:-4]+'.pdf')
                })
        # Phase 2: Render all elements
        removeDocAttributes(root)
        for dirExamples in examples.values():
            for example in dirExamples:
                xml = etree.tounicode(example['element']).strip()
                xml = re.sub(
                    ' ?xmlns:doc="http://namespaces.zope.org/rml/doc"', '', xml)
                xml = dedent(xml)
                xml = enforceColumns(xml, 80)
                xml = highlightRML(xml)
                example['code'] = xml

        rmlFile.close()

    return examples


def main(outPath=None):
    examples = extractExamples(EXAMPLES_DIRECTORY)

    template = pagetemplate.RMLPageTemplateFile('reference.pt')

    directives = {}
    processSignature('document', document.IDocument, examples, directives)
    directives = sorted(directives.values(), key=lambda d: d['name'])

    pdf = template(
        version=__version__,
        types=getAttributeTypes(),
        directives=directives)
    file_ = open(outPath or 'rml-reference.pdf', 'wb')
    file_.write(pdf)
    file_.close()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing all XML Locale functionality.
"""
import logging
import os
import subprocess
import unittest
import sys
import z3c.rml.tests
from z3c.rml import rml2pdf

try:
    import Image
except ImportError:
    from PIL import Image

LOG_FILE = os.path.join(os.path.dirname(__file__), 'render.log')


def gs_command(path):
    return ('gs', '-q', '-sNOPAUSE', '-sDEVICE=png256',
            '-sOutputFile=%s[Page-%%d].png' % path[:-4],
            path, '-c', 'quit')


class RMLRenderingTestCase(unittest.TestCase):

    def __init__(self, inPath, outPath):
        self._inPath = inPath
        self._outPath = outPath
        unittest.TestCase.__init__(self)

    def setUp(self):
        import z3c.rml.tests.module
        sys.modules['module'] = z3c.rml.tests.module
        sys.modules['mymodule'] = z3c.rml.tests.module

    def tearDown(self):
        del sys.modules['module']
        del sys.modules['mymodule']

        from z3c.rml.document import LOGGER_NAME
        for handler in logging.getLogger(LOGGER_NAME).handlers:
            if handler.baseFilename == LOG_FILE:
                handler.close()

        if os.path.exists(LOG_FILE):
            os.remove(LOG_FILE)

    def runTest(self):
        rml2pdf.go(self._inPath, self._outPath)


class ComparePDFTestCase(unittest.TestCase):

    level = 2

    def __init__(self, basePath, testPath):
        self._basePath = basePath
        self._testPath = testPath
        unittest.TestCase.__init__(self)

    def assertSameImage(self, baseImage, testImage):
        base_file = open(baseImage, 'rb')
        test_file = open(testImage, 'rb')
        base_image = Image.open(base_file)
        base = base_image.getdata()
        test_image = Image.open(test_file)
        test = test_image.getdata()
        err_count = 0

        for i in range(len(base)):
            if (base[i] - test[i]) != 0:
                err_count += 1

                if False:
                    from base64 import b64encode
                    import PIL
                    differ = PIL.ImageChops.subtract(base_image, test_image)
                    # differ.show()
                    # output the result as base64 for travis debugging
                    print(b64encode(differ.tobytes()))
                    # flip the above condition to activate this code
                    print()
                    print(os.system("gs --version"))

                    base_file.seek(0)
                    print(baseImage)
                    print(b64encode(base_file.read()))
                    print(self._basePath)
                    with open(self._basePath, 'rb') as base_pdf:
                        print(b64encode(base_pdf.read()))

                    test_file.seek(0)
                    print(testImage)
                    print(b64encode(test_file.read()))
                    print(self._testPath)
                    with open(self._testPath, 'rb') as test_pdf:
                        print(b64encode(test_pdf.read()))

                self.fail(
                    'Image is not the same: %s' % os.path.basename(baseImage))
        base_file.close()
        test_file.close()

    def runTest(self):
        # Convert the base PDF to image(s)
        status = subprocess.Popen(gs_command(self._basePath)).wait()
        if status:
            return
        # Convert the test PDF to image(s)
        status = subprocess.Popen(gs_command(self._testPath)).wait()
        if status:
            return
        # Go through all pages and ensure their equality
        n = 1
        while True:
            baseImage = self._basePath[:-4] + '[Page-%i].png' %n
            testImage = self._testPath[:-4] + '[Page-%i].png' %n
            if os.path.exists(baseImage) and os.path.exists(testImage):
                self.assertSameImage(baseImage, testImage)
            else:
                break
            n += 1


class CompareFileTestCase(unittest.TestCase):

    def __init__(self, testPath, contains):
        self._testPath = testPath
        self._contains = contains
        unittest.TestCase.__init__(self)

    def runTest(self):
        f = open(self._testPath, 'rb')
        try:
            contents = f.read()

            if self._contains not in contents:
                self.fail(
                    'PDF file does not contain: %s' % self._contains
                )
        finally:
            f.close()


def test_suite():
    if False:
        # Debug info
        import pkg_resources
        print("Environment\n===========")
        print("reportlab:")
        print(pkg_resources.get_distribution('reportlab').version)
        print("ghostscript:")
        os.system("gs --version")

    suite = unittest.TestSuite()
    here = os.path.dirname(z3c.rml.tests.__file__)
    inputDir = os.path.join(here, 'input')
    outputDir = os.path.join(here, 'output')
    expectDir = os.path.join(here, 'expected')
    for filename in os.listdir(inputDir):
        if not filename.endswith(".rml"):
            continue

        inPath = os.path.join(inputDir, filename)
        outPath = os.path.join(outputDir, filename[:-4] + '.pdf')
        expectPath = os.path.join(expectDir, filename[:-4] + '.pdf')

        # ** Test RML to PDF rendering **
        # Create new type, so that we can get test matching
        TestCase = type(filename[:-4], (RMLRenderingTestCase,), {})
        case = TestCase(inPath, outPath)
        suite.addTest(case)
        # ** Test PDF rendering correctness **
        TestCase = type('compare-'+filename[:-4], (ComparePDFTestCase,), {})
        case = TestCase(expectPath, outPath)
        suite.addTest(case)

        if filename == 'printScaling.rml':
            TestCase = type('compare-file-'+filename[:-4],
                            (CompareFileTestCase,), {})
            case = TestCase(outPath, b'/PrintScaling /None')
            suite.addTest(case)

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2007-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Paragraph-internal XML parser extensions.
"""
import six
import sys
import reportlab.lib.fonts
import reportlab.lib.styles
import reportlab.lib.utils
import reportlab.platypus.paraparser
import reportlab.platypus.paragraph

class ParaFragWrapper(reportlab.platypus.paraparser.ParaFrag):
    @property
    def text(self):
        key = self._get_pass_key()
        if not hasattr(self, key):
            setattr(self, key, self._get_text())
        return getattr(self, key)

    @text.setter
    def text(self, value):
        key = self._get_pass_key()
        setattr(self, key, value)

    def _get_pass_key(self):
        canvas = self._get_canvas()
        return '_text_%s_%s' % (
            canvas._doctemplate.current_pass, canvas.getPageNumber()
        )

    def _get_canvas(self):
        canvas = None
        i = 5

        while canvas is None:
            try:
                frame = sys._getframe(i)
            except ValueError:
                raise Exception(
                    "Can't use <%s> in this location." % self.__tag__)

            # Guess 1: We're in a paragraph in a story.
            canvas = frame.f_locals.get('canvas', None)

            if canvas is None:
                # Guess 2: We're in a template
                canvas = frame.f_locals.get('canv', None)

            if canvas is None:
                # Guess 3: We're in evalString or namedString
                canvas = getattr(frame.f_locals.get('self', None), 'canv', None)

            i += 1

        return canvas


class PageNumberFragment(ParaFragWrapper):
    """A fragment whose `text` is computed at access time."""

    __tag__ = 'pageNumber'

    def __init__(self, **attributes):
        reportlab.platypus.paraparser.ParaFrag.__init__(self, **attributes)
        self.counting_from = attributes.get('countingFrom', 1)

    def _get_text(self):
        canvas = self._get_canvas()
        return str(canvas.getPageNumber() + int(self.counting_from) - 1)


class GetNameFragment(ParaFragWrapper):
    """A fragment whose `text` is computed at access time."""

    __tag__ = 'getName'

    def __init__(self, **attributes):
        reportlab.platypus.paraparser.ParaFrag.__init__(self, **attributes)
        self.id = attributes['id']
        self.default = attributes.get('default')

    def _get_text(self):
        canvas = self._get_canvas()
        return canvas.manager.get_name(self.id, self.default)


class EvalStringFragment(ParaFragWrapper):
    """A fragment whose `text` is evaluated at access time."""

    __tag__ = 'evalString'

    def __init__(self, **attributes):
        reportlab.platypus.paraparser.ParaFrag.__init__(self, **attributes)
        self.frags = attributes.get('frags', [])

    def _get_text(self):
        text = u''
        for frag in self.frags:
            if isinstance(frag, six.string_types):
                text += frag
            else:
                text += frag.text
        from z3c.rml.special import do_eval
        return do_eval(text)


class NameFragment(ParaFragWrapper):
    """A fragment whose attribute `value` is set to a variable."""

    __tag__ = 'name'

    def __init__(self, **attributes):
        reportlab.platypus.paraparser.ParaFrag.__init__(self, **attributes)

    def _get_text(self):
        canvas = self._get_canvas()
        canvas.manager.names[self.id] = self.value
        return u''


class SpanStyle(reportlab.lib.styles.PropertySet):
    # Do not specify defaults, so that all attributes can be inherited.
    defaults = {}


class Z3CParagraphParser(reportlab.platypus.paraparser.ParaParser):
    """Extensions to paragraph-internal XML parsing."""

    def __init__(self, manager, *args, **kwargs):
        reportlab.platypus.paraparser.ParaParser.__init__(self, *args, **kwargs)
        self.manager = manager
        self.in_eval = False

    def findSpanStyle(self, style):
        from z3c.rml import attr
        return attr._getStyle(self.manager, style)

    def startDynamic(self, attributes, klass):
        frag = klass(**attributes)
        frag.__dict__.update(self._stack[-1].__dict__)
        frag.__tag__ = klass.__tag__
        frag.fontName = reportlab.lib.fonts.tt2ps(
            frag.fontName, frag.bold, frag.italic)
        if self.in_eval:
            self._stack[-1].frags.append(frag)
        else:
            self.fragList.append(frag)
            self._stack.append(frag)

    def endDynamic(self):
        if not self.in_eval:
            self._stack.pop()

    def start_pagenumber(self, attributes):
        self.startDynamic(attributes, PageNumberFragment)

    def end_pagenumber(self):
        self.endDynamic()

    def start_getname(self, attributes):
        self.startDynamic(attributes, GetNameFragment)

    def end_getname(self):
        self.endDynamic()

    def start_name(self, attributes):
        self.startDynamic(attributes, NameFragment)

    def end_name(self):
        self.endDynamic()

    def start_evalstring(self, attributes):
        self.startDynamic(attributes, EvalStringFragment)
        self.in_eval = True

    def end_evalstring(self):
        self.in_eval = False
        self.endDynamic()

    def handle_data(self, data):
        if not self.in_eval:
            reportlab.platypus.paraparser.ParaParser.handle_data(self, data)
        else:
            frag = self._stack[-1].frags.append(data)


class Z3CParagraph(reportlab.platypus.paragraph.Paragraph):
    """Support for custom paraparser with sytles knowledge.

    Methods mostly copied from reportlab.
    """

    def __init__(self, text, style, bulletText = None, frags=None,
                 caseSensitive=1, encoding='utf8', manager=None):
        self.caseSensitive = caseSensitive
        self.encoding = encoding
        self._setup(
            text, style, bulletText or getattr(style,'bulletText',None), frags,
            reportlab.platypus.paragraph.cleanBlockQuotedText, manager)

    def _setup(self, text, style, bulletText, frags, cleaner, manager):

        # This used to be a global parser to save overhead.  In the interests
        # of thread safety it is being instantiated per paragraph.  On the
        # next release, we'll replace with a cElementTree parser

        if frags is None:
            text = cleaner(text)
            _parser = Z3CParagraphParser(manager)
            _parser.caseSensitive = self.caseSensitive
            style, frags, bulletTextFrags = _parser.parse(text, style)
            if frags is None:
                raise ValueError(
                    "xml parser error (%s) in paragraph beginning\n'%s'"\
                    % (_parser.errors[0],text[:min(30,len(text))]))
            reportlab.platypus.paragraph.textTransformFrags(frags,style)
            if bulletTextFrags:
                bulletText = bulletTextFrags

        # AR hack
        self.text = text
        self.frags = frags  # either the parse fragments or frag word list
        self.style = style
        self.bulletText = bulletText
        self.debug = 0


class ImageReader(reportlab.lib.utils.ImageReader):

    def __init__(self, fileName, ident=None):
        if isinstance(fileName, six.string_types):
            # Avoid circular imports. The filename resolutions hould be added
            # to a utility module.
            from z3c.rml import attr
            _srcAttr = attr.File(doNotOpen=True)
            fileName = _srcAttr.fromUnicode(fileName)
        return super(ImageReader, self).__init__(fileName, ident)


# Monkey Patch reportlab image reader, so that <img> tags inside paragraphs
# can also benefit from package-local file paths.
reportlab.platypus.paraparser.ImageReader = ImageReader



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing all XML Locale functionality.
"""
import os
import unittest
import z3c.rml.tests
from z3c.rml import rml2pdfscript
from z3c.rml.tests.test_rml import ComparePDFTestCase
from z3c.rml.tests.test_rml import RMLRenderingTestCase


class RMLRenderingTestCase(RMLRenderingTestCase):

    def runTest(self):
        rml2pdfscript.goSubProcess(self._inPath, self._outPath, True)


def test_suite():
    suite = unittest.TestSuite()
    if False:
        inputDir = os.path.join(os.path.dirname(z3c.rml.tests.__file__), 'input')
        outputDir = os.path.join(os.path.dirname(z3c.rml.tests.__file__), 'output')
        expectDir = os.path.join(os.path.dirname(z3c.rml.tests.__file__), 'expected')
        for filename in os.listdir(inputDir):
            if not filename.endswith(".rml"):
                continue
            inPath = os.path.join(inputDir, filename)
            outPath = os.path.join(outputDir, filename[:-4] + '.pdf')
            expectPath = os.path.join(expectDir, filename[:-4] + '.pdf')

            # ** Test RML to PDF rendering **
            # Create new type, so that we can get test matching
            TestCase = type(filename[:-4], (RMLRenderingTestCase,), {})
            case = TestCase(inPath, outPath)
            suite.addTest(case)

            # ** Test PDF rendering correctness **
            TestCase = type('compare-'+filename[:-4], (ComparePDFTestCase,), {})
            case = TestCase(expectPath, outPath)
            suite.addTest(case)

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``pdfInclude`` Directive.
"""
__docformat__ = "reStructuredText"
from reportlab.platypus import flowables

from z3c.rml import attr, flowable, interfaces, occurence


class StoryPlaceFlowable(flowables.Flowable):

    def __init__(self, x, y, width, height, origin, flows):
        flowables.Flowable.__init__(self)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.origin = origin
        self.flows = flows

    def wrap(self, *args):
        return (0, 0)

    def draw(self):
        saveState = False
        x, y = self.x, self.y
        self.canv.restoreState()
        if self.origin == 'frame':
            x += self._frame._x1
            y += self._frame._y1
        elif self.origin == 'local':
            x += self._frame._x
            y += self._frame._y
        else:
            # origin == 'page'
            pass
        width, height  = self.width, self.height
        y += height
        for flow in self.flows.flow:
            flowWidth, flowHeight = flow.wrap(width, height)
            if flowWidth <= width and flowHeight <= height:
                y -= flowHeight
                flow.drawOn(self.canv, x, y)
                height -= flowHeight
            else:
                raise ValueError("Not enough space")
        self.canv.saveState()


class IStoryPlace(interfaces.IRMLDirectiveSignature):
    """Draws a set of flowables on the canvas within a given region."""

    x = attr.Measurement(
        title=u'X-Coordinate',
        description=(u'The X-coordinate of the lower-left position of the '
                     u'place.'),
        required=True)

    y = attr.Measurement(
        title=u'Y-Coordinate',
        description=(u'The Y-coordinate of the lower-left position of the '
                     u'place.'),
        required=True)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the place.',
        required=False)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the place.',
        required=False)

    origin = attr.Choice(
        title=u'Origin',
        description=u'The origin of the coordinate system for the story.',
        choices=('page', 'frame', 'local'),
        default = 'page',
        required=False)

class StoryPlace(flowable.Flowable):
    signature = IStoryPlace

    def process(self):
        x, y, width, height, origin = self.getAttributeValues(
            select=('x', 'y', 'width', 'height', 'origin'), valuesOnly=True)

        flows = flowable.Flow(self.element, self.parent)
        flows.process()
        self.parent.flow.append(
            StoryPlaceFlowable(x, y, width, height, origin, flows))


flowable.Flow.factories['storyPlace'] = StoryPlace
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('storyPlace', IStoryPlace),)
    )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Special Element Processing
"""
import six
from z3c.rml import attr, directive, interfaces


class IName(interfaces.IRMLDirectiveSignature):
    """Defines a name for a string."""

    id = attr.String(
        title=u'Id',
        description=u'The id under which the value will be known.',
        required=True)

    value = attr.Text(
        title=u'Value',
        description=u'The text that is displayed if the id is called.',
        required=True)

class Name(directive.RMLDirective):
    signature = IName

    def process(self):
        id, value = self.getAttributeValues(valuesOnly=True)
        manager = attr.getManager(self)
        manager.names[id] = value


class IAlias(interfaces.IRMLDirectiveSignature):
    """Defines an alias for a given style."""

    id = attr.String(
        title=u'Id',
        description=u'The id as which the style will be known.',
        required=True)

    value = attr.Style(
        title=u'Value',
        description=u'The style that is represented.',
        required=True)

class Alias(directive.RMLDirective):
    signature = IAlias

    def process(self):
        id, value = self.getAttributeValues(valuesOnly=True)
        manager = attr.getManager(self)
        manager.styles[id] = value


class TextFlowables(object):
    def _getManager(self):
        if hasattr(self, 'manager'):
            return self.manager
        else:
            return  attr.getManager(self)

    def getPageNumber(self, elem, canvas):
        return six.text_type(
            canvas.getPageNumber() + int(elem.get('countingFrom', 1)) - 1
        )

    def getName(self, elem, canvas):
        return self._getManager().get_name(
            elem.get('id'),
            elem.get('default')
        )

    def evalString(self, elem, canvas):
        return do_eval(self._getText(elem, canvas, False))

    def namedString(self, elem, canvas):
        self._getManager().names[elem.get('id')] = self._getText(
            elem, canvas, include_final_tail=False
        )
        return u''

    def name(self, elem, canvas):
        self._getManager().names[elem.get('id')] = elem.get('value')
        return u''

    handleElements = {'pageNumber': getPageNumber,
                      'getName': getName,
                      'evalString': evalString,
                      'namedString': namedString,
                      'name': name}

    def _getText(self, node, canvas, include_final_tail=True):
        text = node.text or u''
        for sub in node.getchildren():
            if sub.tag in self.handleElements:
                text += self.handleElements[sub.tag](self, sub, canvas)
            else:
                self._getText(sub, canvas)
            text += sub.tail or u''
        if include_final_tail:
            text += node.tail or u''
        return text

def do_eval(value):
    # Maybe still not safe
    value = value.strip()
    if value:
        return six.text_type(eval(value.strip(), {'__builtins__': None}, {}))
    return u''
